// 
// Finding the closest point using 2 leveled subdivision

// University of Maribor
// Faculty of Electrical Engineering & Computer Science
// Laboratory for Geometric Modelling & Multimedia Algorithms - GEMMA
// Written By Borut ZALIK in August 2000
// e-mail: zalik@uni-mb.si
//////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "Proximity2L.h"
#include "utility.h"
#include "stdio.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//#define PRINT_TO_F1

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

static int Threshold = 0;




DCL::DCL(double l, double b, double r, double u, double sx, double sy)
// initialisation of one cells:
// l, b, r, u: values of left, bottom, right and uppr border of the cell
// sx, sy: size of the cell
{ top = NULL;
  SecondLevelArray = NULL; 
  NoOfPoints = 0;
  Left = l; Bottom = b;
  Right = r; Upper = u;
  SizeSmallX = sx; 
  SizeSmallY = sy;
  VisitedFlag = false;
}



void DCL::DeleteList()
// one level list of points is erased
{ DCList* temp;
  while (top != NULL)
  { temp = top;
    top = top->Next;
    delete temp;
  }
}


DCL::~DCL()
// free memory of the list of points storred at one cell
{ if (top != NULL)
    DeleteList();  
}



void DCL::PlotGrid(HDC hdc, double a1, double b1,
                            double a2, double b2,
                            int rasterX, int rasterY)
// plot the grid of the second level
{ int i, j;
  double x1, y1, x2, y2;
  int xx1, yy1, xx2, yy2;


 
  for (i = 0; i < NO_OF_SMALL_CELLS; i++)
    for (j = 0; j < NO_OF_SMALL_CELLS; j++)   
    { x1 = Left + i * SizeSmallX;
      y1 = Bottom + j * SizeSmallY;
      x2 = x1 + SizeSmallX;
      y2 = y1 + SizeSmallY;


      xx1 = FromRealToPixel(x1, a1, a2, rasterX);
      yy1 = FromRealToPixel(y1, b1, b2, rasterY);
      xx2 = FromRealToPixel(x2, a1, a2, rasterX);
      yy2 = FromRealToPixel(y1, b1, b2, rasterY);

      MoveToEx(hdc, xx1, yy1, NULL);
      LineTo(hdc, xx2, yy2);

      xx2 = FromRealToPixel(x2, a1, a2, rasterX);
      yy2 = FromRealToPixel(y2, b1, b2, rasterY);
      LineTo(hdc, xx2, yy2);

      xx2 = FromRealToPixel(x1, a1, a2, rasterX);
      yy2 = FromRealToPixel(y2, b1, b2, rasterY);
      LineTo(hdc, xx2, yy2);
      LineTo(hdc, xx1, yy1);
    }
}




void DCL::InsertPoint(MyPoint* p)
// inserts the point which is pointed by p into two leveled subdivision
{ if (NoOfPoints < Threshold) // we are inserting in the first level
  { NoOfPoints++;
    AddPoint(p);
  }
  else
  { if (SecondLevelArray == NULL)
    {  // CreateSecondLevel 
       // memory is alocated
      SecondLevelArray = new DCL**[NO_OF_SMALL_CELLS];
      for (int i = 0; i < NO_OF_SMALL_CELLS; i++) 
        SecondLevelArray[i] = new DCL*[NO_OF_SMALL_CELLS];

      int j;
      double l, b;
      // cells are initialised

      l = Left - SizeSmallX;
      for (i = 0; i < NO_OF_SMALL_CELLS; i++)
      { l += SizeSmallX;
        b = Bottom - SizeSmallY;
        for (j = 0; j < NO_OF_SMALL_CELLS; j++)
        { b += SizeSmallY;          
          SecondLevelArray[i][j] = new DCL(l, b, l + SizeSmallX, b + SizeSmallY, -1, -1);
        }  
      } 

      // old points are inserted into second level
/*      DCList* temp;
      MyPoint* qq;
      int px, py;
      double pxx, pyy;
      while (top != NULL)
      { temp = top;
        qq = top->ReturnPoint();  
        qq->ReturnXY(pxx, pyy);
 
        px = (int)((pxx - Left)/SizeSmallX);
        py = (int)((pyy - Bottom)/SizeSmallY);
        SecondLevelArray[px][py]->AddPoint(qq);    
        top = top->Next;
        delete temp;
      }

      top  = NULL;  
*/

    }
  

   
    double pvx, pvy;
    p->ReturnXY(pvx, pvy);
    int PosX = (int)((pvx - Left)/SizeSmallX);
    int PosY = (int)((pvy - Bottom)/SizeSmallY);

    SecondLevelArray[PosX][PosY]->AddPoint(p);      
  }  
}


void DCL::SecondLevelPosition(MyPoint* p, int& PosX, int& PosY)
// determines the position of point p in the second level and returns that coordinates
{
  double x, y;
  p->ReturnXY(x, y);
  PosX = (int)((x - Left)/SizeSmallX);
  PosY = (int)((y - Bottom)/SizeSmallY);
}

void DCL::AddPoint(MyPoint* a)
// let us add point with index "index" into the list of points at considered cell
{ 
  DCList* temp = new DCList(a);
  temp->Next = top;
  top = temp;
}


void DCL::AddPoint(DCList* temp)
{
  temp->Next = top;
  top = temp;
}


Proximity2L::Proximity2L(double ax1, double ay1, double bx1, double by1, int n)
// prepared 2-leveled uniform grid for the use
// first 4 parameters are bounding square coordinates
// n is total number of points
{ 
  spiral = NULL; 
  TheNextClosestPointList = NULL;

  AuxDCLList = NULL;

  if (n == 0)
    dclArray = NULL;
  else
  { int i, j;
    
    double a = 0.02 * (bx1 - ax1);
    double b = 0.02 * (by1 - ay1);
  
    BoxXmin = ax1 - a; // bounding box is increased a bit
    BoxYmin = ay1 - b;
    BoxXmax = bx1 + a;
    BoxYmax = by1 + b;
      
    double deltaX = BoxXmax - BoxXmin;
    double deltaY = BoxYmax - BoxYmin;
    double ratioXY = deltaX/deltaY;

    
    // determine number of cells at the first level using some heuristic
    NoOfCellsX1 = (int)(0.5 * sqrt(n) * ratioXY  );
    NoOfCellsY1 = (int)(0.5 * sqrt(n) / ratioXY  );

    // too much cells
    if (NoOfCellsX1 > UPPER_ARRAY_LIMIT)
      NoOfCellsX1 = UPPER_ARRAY_LIMIT;
     
    if (NoOfCellsY1 > UPPER_ARRAY_LIMIT)
      NoOfCellsY1 = UPPER_ARRAY_LIMIT;
 
   // not enough cells

    if (NoOfCellsX1 < 1)
      NoOfCellsX1 = 1;
     
    if (NoOfCellsY1 < 1)
      NoOfCellsY1 = 1;

/*
    NoOfCellsX1 = (int)(sqrt(n) * ratioXY  );
    NoOfCellsY1 = (int)(sqrt(n) / ratioXY  );
NoOfCellsX1 = (int)(NoOfCellsX1 / 4);
NoOfCellsY1 = (int)(NoOfCellsY1 / 4);
*/
    BigDiagonal = NoOfCellsX1*NoOfCellsY1 + 10;


    // determine size of cells at the first level
    XSize1 = deltaX / NoOfCellsX1;
    YSize1 = deltaY / NoOfCellsY1;

    // determine number of points in the cells which couses the second level is introduced
//    Threshold = (int)(2 * n/(NoOfCellsX1 * NoOfCellsY1) );

//Threshold = (int)(4 * n/(NoOfCellsX1 * NoOfCellsY1) );

 Threshold = 100000;

//Threshold = 12;

/*
    // there is too many cells
    if (NoOfCellsX1 > UPPER_ARRAY_LIMIT)
    { NoOfCellsX1 = UPPER_ARRAY_LIMIT;
      XSize1 = deltaX / UPPER_ARRAY_LIMIT;
    }


    if (NoOfCellsY1 > UPPER_ARRAY_LIMIT)
    { NoOfCellsY1 = UPPER_ARRAY_LIMIT;
      YSize1 = deltaY / UPPER_ARRAY_LIMIT;
    }
*/

    // the memory is allocated for the first level
    dclArray = new DCL**[NoOfCellsX1];
    for (i = 0; i < NoOfCellsX1; i++)
      dclArray[i] = new DCL*[NoOfCellsY1];

    
    double xa1 = XSize1 / NO_OF_SMALL_CELLS;
    double ya1 = YSize1 / NO_OF_SMALL_CELLS;
    double le, bo;

    for (i = 0; i < NoOfCellsX1; i++)
      for (j = 0; j < NoOfCellsY1; j++)
      { le = BoxXmin + i * XSize1;
        bo = BoxYmin + j * YSize1;
        dclArray[i][j] = new DCL(le, bo, le + XSize1, bo + YSize1,xa1, ya1);
      }
  }
}


Proximity2L::~Proximity2L()
// free memory occupied by the two-leveled uniform subdivision
{  
  DeleteSpiral();


  if (dclArray != NULL)
  { int i, j, kk, ll;
    DCL*** SLA;


    for (i = 0; i < NoOfCellsX1; i++)
      for (j = 0; j < NoOfCellsY1; j++)
      { 
        SLA = dclArray[i][j]->ReturnSecondLevel();
        
        if ( SLA != NULL)
        { 
          for (kk = 0; kk < NO_OF_SMALL_CELLS; kk++)
            for (ll = 0; ll < NO_OF_SMALL_CELLS; ll++)
              if (SLA[kk] != NULL) 
                delete SLA[kk][ll];       

          for (kk = 0; kk < NO_OF_SMALL_CELLS; kk++)
            delete[] SLA[kk];

          delete[] SLA;
        }
 
        delete dclArray[i][j]; 
      } 

    for (i = 0; i < NoOfCellsX1; i++)
      delete[] dclArray[i];

     delete[] dclArray;

  }


  DCList* temp;

  while (TheNextClosestPointList != NULL)
  { temp = TheNextClosestPointList;
    TheNextClosestPointList = TheNextClosestPointList->Next;
    delete temp;
  }


  if (AuxDCLList != NULL)
  { DCList* a;
    MyPoint* p;

    while (AuxDCLList != NULL)
    { p = AuxDCLList->ReturnPoint();
      a = AuxDCLList;
      AuxDCLList = AuxDCLList->Next;
      delete a;
      delete p;
    }

  }

}


void Proximity2L::DeleteSpiral()
// delete one level list 
{  
 SpiralClass* stemp;
 
 while (spiral != NULL)
 { stemp = spiral;
   spiral = spiral->Next;
   delete stemp;
 }
}


void Proximity2L::DeleteTheNextClosestPointList()
{ DCList* temp;
  MyPoint* mp;

  while (TheNextClosestPointList != NULL)
  { temp = TheNextClosestPointList;
    TheNextClosestPointList = TheNextClosestPointList->Next;
    mp = temp->ReturnPoint();
    if (mp != NULL)
       mp->ResetNextCloseFlag();
    delete temp;
  }
}



void Proximity2L::AddAndMarkTheNextClosestPointList(MyPoint* q)
{ DCList* qq = new DCList(q);

  q->SetNextCloseFlag();
  qq->Next = TheNextClosestPointList;
  TheNextClosestPointList = qq;
}



void Proximity2L::DeleteSecondLevel(DCL*** SLA)
// free memory occupied by the second level
{ int i,j;

  for (i = 0; i < NO_OF_SMALL_CELLS; i++)
    for (j =0; j < NO_OF_SMALL_CELLS; j++)
      delete SLA[i][j];

  for (i = 0; i < NO_OF_SMALL_CELLS; i++)
    delete[] SLA[i];

  delete[] SLA;  
}


void Proximity2L::FirstLevelPosition(MyPoint* p, int& PosX, int& PosY)
// method determines the coordinate of the cell into which point p falls
{ double x, y;
  p->ReturnXY(x, y);
  PosX = (int)((x - BoxXmin)/XSize1);
  PosY = (int)((y - BoxYmin)/YSize1);
}

void Proximity2L::InsertPoint(MyPoint* p)
// method inserts the point p with index "index" into the 2 leveled uniform subdivision
{ int PosX, PosY;
  FirstLevelPosition(p, PosX, PosY);
  dclArray[PosX][PosY]->InsertPoint(p); 
}

void Proximity2L::InsertPoint(double xx, double yy, long ind)
{ MyPoint* p = new MyPoint(xx, yy, ind);
  p->ResetNextCloseFlag();
  DCList *a = new DCList(p);
   
  a->Next = AuxDCLList;
  AuxDCLList = a;

  InsertPoint(p);
  
}


void Proximity2L::PlotGrid(HDC hdc, int right, int bottom)
// auxiliary method. I am not sure, if it works completely correct
{ int i, j;
  double x1, y1, x2, y2;
  int xx1, yy1, xx2, yy2;

//  double deltax = BOXDELTA * (BoxXmax - BoxXmin);
//  double deltay = BOXDELTA * (BoxYmax - BoxYmin);
 double deltax = 0;
 double deltay = 0;

  for (i = 0; i < NoOfCellsX1; i++)
    for (j = 0; j < NoOfCellsY1; j++)   
    { x1 = BoxXmin + i * XSize1;
      y1 = BoxYmin + j * YSize1;
      x2 = x1 + XSize1;
      y2 = y1 + YSize1;

      xx1 = FromRealToPixel(x1, BoxXmin - deltax, BoxXmax-BoxXmin + 2*deltax, right);
      yy1 = FromRealToPixel(y1, BoxYmin - deltay, BoxYmax-BoxYmin + 2*deltay, bottom);
      xx2 = FromRealToPixel(x2, BoxXmin - deltax, BoxXmax-BoxXmin + 2*deltax, right);
      yy2 = FromRealToPixel(y1, BoxYmin - deltay, BoxYmax-BoxYmin + 2*deltay, bottom);



      MoveToEx(hdc, xx1, yy1, NULL);
      LineTo(hdc, xx2, yy2);

      xx2 = FromRealToPixel(x2, BoxXmin - deltax, BoxXmax-BoxXmin + 2*deltax, right);
      yy2 = FromRealToPixel(y2, BoxYmin - deltay, BoxYmax-BoxYmin + 2*deltay, bottom);
      LineTo(hdc, xx2, yy2);

      xx2 = FromRealToPixel(x1, BoxXmin - deltax, BoxXmax-BoxXmin+ 2*deltax, right);
      yy2 = FromRealToPixel(y2, BoxYmin - deltay, BoxYmax-BoxYmin+ 2*deltay, bottom);
      LineTo(hdc, xx2, yy2);
      LineTo(hdc, xx1, yy1);
    }

  
  for (i = 0; i < NoOfCellsX1; i++)
    for (j = 0; j < NoOfCellsY1; j++)   
      if (dclArray[i][j]->ReturnSecondLevel() != NULL)
        dclArray[i][j]->PlotGrid(hdc, BoxXmin - deltax, BoxYmin - deltay, 
                                      BoxXmax-BoxXmin + 2*deltax, 
                                      BoxYmax-BoxYmin + 2*deltay, 
                                      right, bottom);

}



DCList* Proximity2L::FindTheClosestPointInCell(DCList* list, double& d)
// Method goes through the list of points and determines the closest point to the ReferencePoint
// - list: pointer to the list of points
// - current distance which is updated if neccesary
// - true if the point with the same coordinate exist
//  Returns the closestPoint, or NULL, if there is no such point, or points
//  with the same coordinates exist
{ 
  DCList* TheClosestPointRecord = NULL;
  double d1;
  double x,y;
  

  while (list != NULL)
  { if (! list->ReturnPoint()->ReturnNextCloseFlag())
    { list->ReturnPointCoordinate(x, y);
      d1 = Sqr(ReferencePoint.x - x) + Sqr(ReferencePoint.y - y);
      if (d1 < d)
      { d = d1;
        TheClosestPointRecord = list;
      }
    }
    list = list->Next;
  } 

  return TheClosestPointRecord;
}


DCList* Proximity2L::FindTheClosestPointInCell2Level(
                          DCL*** Array2,
                          int ii, int jj, double& distance)
// The closest point to the ReferencePoint is found if current cell is divided.
// - Array2: the seond level array
// - ii, jj: coordinates of the cells where the search is started
// - distance: the smallest distance found up to know. If there exist closer point,
//             distance is updated and pointer to that point is returned. 
{ DCList *PPointer, *PPointer1;
  double d1, d2;
  
  PPointer = NULL;
  
  PPointer = FindTheClosestPointInCell(Array2[ii][jj]->ReturnTopOfList(), distance);
  int i, j;
  double x, y;

  for (i = 0; i < NO_OF_SMALL_CELLS; i++)    // let us visit all cells at the second level
    for (j = 0; j < NO_OF_SMALL_CELLS; j++)
    { if (!((i == ii) && (j == jj)))         // cell (ii, jj) has already been checked
      { if (i == ii)                         // this cell has the same x index
        { if (j < jj)                        // if it is below the cell ii
            Array2[i][j]->ReturnRU(x, y);    // let us return upper border of that cell
          else
          {  if (j > jj)                     //  the cell is upper the cells ii
              Array2[i][j]->ReturnRB(x, y);  // we need the bottom border of that cells
          } 
          x = ReferencePoint.x;              // x coordinate returned is the same as ReferencePoint
          goto ven;                          // other checks are not needed
        }

        if (j == jj)                         // simmilar as before, but this time the cell has  
        { if (i < ii)                        // y coordinate
            Array2[i][j]->ReturnRU(x, y);
          else
          { if ( i > ii)
              Array2[i][j]->ReturnLU(x, y);
          }
          y = ReferencePoint.y; 
          goto ven;
        } 

        if ( i < ii)                         // cells left to the cell (ii, jj) are examined
        { if (j < jj)
            Array2[i][j]->ReturnRU(x, y);
          else
            Array2[i][j]->ReturnRB(x, y);
          goto ven;
        }

       if ( i > ii)                           //  cells right to the cells (ii, jj) are examined
        { if (j < jj)
            Array2[i][j]->ReturnLU(x, y);
          else
            Array2[i][j]->ReturnLB(x, y);

        }

       // let us determine now, how fare the most faraway cell from Reference point is 
ven:   d1 = Sqr(ReferencePoint.x - x) + Sqr(ReferencePoint.y - y);

       if (d1 < distance) // if cell is closer that the point being found up to know
       { d2 = distance;  
         // let us find the closest point in that cell to the ReferencePoint 
         PPointer1 = FindTheClosestPointInCell(Array2[i][j]->ReturnTopOfList(), d2);
         if (d2 < distance) // if it is closer than the closest point up to know, let us remember it
         { PPointer = PPointer1;
           distance = d2;
         }
       }
      }   
    }
  return PPointer; // the closest point is returned
}


DCList*  Proximity2L::GetDistanceFromSpiral(double& CurrentDistance)
// Method determines if any point storred in the spiral is closer to the ReferencePoint
// than CurrentDistance. If it is, CurrentDistance if updated and the pointer to that point is returned
{ 
  if (spiral != NULL)
  { 
    DCList* temp = NULL;
    double TempD = MAX_DOUBLE;
    
    SpiralClass* list = spiral;   
    DCList *dc, * Temp1;
    
    while ( list != NULL) // let us go through all cells beging located in the spiral
    { dc = list->ReturnList();
      Temp1  = FindTheClosestPointInCell(dc,TempD); 
      if  (TempD < CurrentDistance)
      { CurrentDistance = TempD;
        temp = Temp1;  
      }  
      list = list->Next;
    }
    return temp; 
  }
  return NULL;
}



bool Proximity2L::ConstructSpiral(int level, MyPoint p, int i1, int j1, double distance)
// methods returns all the cells which are for level cells from the cell (i1 j1).
// If any of that cell has distance greater than "distance", it is not inserted.
// At first , the closest cells located at the x+/-level and y+/-level distance
// are estimated and if they are closer than "distance". all remains cells are checked.

{ int counter = 0;
  double x, y;
  int i, j, k;
  DCL*** SLevelPointer = NULL;


  bool NoSense = false; // stays false if all cells are further than CurrentDistance;

  if ( (i = i1-level) >= 0)   // left cells
  { dclArray[i][j1]->ReturnRU(x, y);
    if (distance > Sqr(p.x - x)) // yes include that cells and check all remaining left cells 
    { NoSense = true;
      if ((SLevelPointer = dclArray[i][j1]->ReturnSecondLevel()) != NULL)
        counter = AddSecondLevelCells(0, SLevelPointer, p, distance, counter);
      else  
        counter = AddToSpiral(dclArray[i][j1]->ReturnTopOfList(), counter);

      for (k = j1-level; k <= j1+level; k++)
      { if (k != j1)  // this cell has been checked already
        { if ((k >= 0) && (k < NoOfCellsY1))
          { if (k < j1)
              dclArray[i][k]->ReturnRU(x, y);
        
            if (k > j1)
              dclArray[i][k]->ReturnRB(x, y);
      
            x = Sqr(p.x  - x); 
            y = Sqr(p.y - y); 
            if (distance > x + y) 
            { if ((SLevelPointer = dclArray[i][k]->ReturnSecondLevel()) != NULL)
                counter = AddSecondLevelCells(0, SLevelPointer, p, distance, counter);
              else
                counter = AddToSpiral(dclArray[i][k]->ReturnTopOfList(), counter);    

            }
          } 
        }
      }
    }
  }

  if ((i = i1+level) < NoOfCellsX1)   // right cells
  { dclArray[i][j1]->ReturnLU(x, y);
    if (distance > Sqr(p.x - x)) // yes include that cells and check all remaining left cells 
    { NoSense = true;
      if ((SLevelPointer = dclArray[i][j1]->ReturnSecondLevel()) != NULL)
        counter = AddSecondLevelCells(1, SLevelPointer, p, distance, counter);
      else 
        counter = AddToSpiral(dclArray[i][j1]->ReturnTopOfList(), counter);

      for (k = j1-level; k <= j1+level; k++)
      { if (k != j1)  // this cell has been checked already
        { if ((k >= 0) && (k < NoOfCellsY1))
          { if (k < j1)
              dclArray[i][k]->ReturnLU(x, y);
        
            if (k > j1)
              dclArray[i][k]->ReturnLB(x, y);
      
            x = Sqr(p.x  - x); 
            y = Sqr(p.y - y); 
            if (distance > x + y) 
            { if ((SLevelPointer = dclArray[i][k]->ReturnSecondLevel()) != NULL)
                counter = AddSecondLevelCells(1, SLevelPointer, p, distance, counter);
              else 
                counter = AddToSpiral(dclArray[i][k]->ReturnTopOfList(), counter);

            }
          } 
        }
      }
    }
  }


  if ((j = j1-level) >= 0)   // lower cells
  { dclArray[i1][j]->ReturnLU(x, y);
    if (distance > Sqr(p.y - y)) // yes include that cells and check all remaining left cells 
    { NoSense = true;
      if ((SLevelPointer = dclArray[i1][j]->ReturnSecondLevel()) != NULL)
        counter = AddSecondLevelCells(2, SLevelPointer, p, distance, counter);
      else 
        counter = AddToSpiral(dclArray[i1][j]->ReturnTopOfList(), counter);
   

      for (k = i1-level+1; k <= i1+level-1; k++)
      { if (k != i1)  // this cell has been checked already
        { if ((k >= 0) && (k < NoOfCellsX1))
          { if (k < i1)
              dclArray[k][j]->ReturnRU(x, y);
        
            if (k > i1)
              dclArray[k][j]->ReturnLU(x, y);
      
            x = Sqr(p.x  - x); 
            y = Sqr(p.y - y); 
            if (distance > x + y) 
            { if ((SLevelPointer = dclArray[k][j]->ReturnSecondLevel()) != NULL)
                counter = AddSecondLevelCells(2, SLevelPointer, p, distance, counter);
              else 
                counter = AddToSpiral(dclArray[k][j]->ReturnTopOfList(), counter);

            }
          } 
        }
      }
    }
  }

  if ((j = j1+level) < NoOfCellsY1)   // upper cells
  { dclArray[i1][j]->ReturnLB(x, y);
    if (distance > Sqr(p.y - y)) // yes include that cells and check all remaining left cells 
    { NoSense = true;
      if ((SLevelPointer = dclArray[i1][j]->ReturnSecondLevel()) != NULL)
        counter = AddSecondLevelCells(3, SLevelPointer, p, distance, counter);
      else 
        counter = AddToSpiral(dclArray[i1][j]->ReturnTopOfList(), counter);


      for (k = i1-level+1; k <= i1+level-1; k++)
      { if (k != i1)  // this cell has been checked already
        { if ((k >= 0) && (k < NoOfCellsX1))
          { if (k < i1)
              dclArray[k][j]->ReturnRB(x, y);
        
            if (k > i1)
              dclArray[k][j]->ReturnLB(x, y);
      
            x = Sqr(p.x  - x); 
            y = Sqr(p.y - y); 
            if (distance > x + y) 
            { if ((SLevelPointer = dclArray[k][j]->ReturnSecondLevel()) != NULL)
                counter = AddSecondLevelCells(3, SLevelPointer, p, distance, counter);
              else 
                counter = AddToSpiral(dclArray[k][j]->ReturnTopOfList(), counter);
            }
          } 
        }
      }
    }
  }
  return NoSense;
}

int Proximity2L::AddToSpiral(DCList* list, int counter)
{ if (list != NULL)
  { SpiralClass* s = new SpiralClass(list);
    s->Next = spiral;
    spiral = s;
    counter++;
   }
   return counter;  
}


int Proximity2L::AddSecondLevelCells(int direction, DCL*** SLevelPointer, MyPoint& p, 
                                 double distance, int& counter)

// direcition: 0 - inserts cells from right to left
//             1 - inserts cells from left to right
//             2 - inserts cells from top to bottom
//             3 - inserts cells from bottom to top
// SLevelPointer : second uniform subdivision
// p: tested point
// distance: the closest distance up to now
// counter: number of inserted cells in the spiral
{ int i, j;
  bool ContFlag = true;  
  double x, y;
  int no = 0;

  switch (direction)
  { case 0: 
    { i = NO_OF_SMALL_CELLS - 1;      
      while ((i >= 0) && ContFlag)
      { SLevelPointer[i][0]->ReturnRU(x, y); // if the whole column is close enough
        if (ABS(x - p.x) < distance)
        { for (j = 0; j < NO_OF_SMALL_CELLS; j++)
            counter = AddToSpiral(SLevelPointer[i][j]->ReturnTopOfList(), counter);
          
          i--;
        }  
        else
          ContFlag = false;
      }
      break;
    }  

    case 1:
    { i = 0;     
      while ((i < NO_OF_SMALL_CELLS) && ContFlag)
      { SLevelPointer[i][0]->ReturnLU(x, y);  
        if (ABS(x - p.x) < distance)
        { for (j = 0; j < NO_OF_SMALL_CELLS; j++)
            counter = AddToSpiral(SLevelPointer[i][j]->ReturnTopOfList(), counter);
          
          i++;
        }  
        else
          ContFlag = false;
      }
      break;
    }  

    case 2:
    { j = NO_OF_SMALL_CELLS -1;      
      while ((j >= 0) && ContFlag)
      { SLevelPointer[0][j]->ReturnRU(x, y); 
        if (ABS(y - p.y) < distance)
        { for (i = 0; i < NO_OF_SMALL_CELLS; i++)
            counter = AddToSpiral(SLevelPointer[i][j]->ReturnTopOfList(), counter);
          j--;
        }
        else
          ContFlag = false;
      }
      break;
    }  

    case 3:
    { j = 0;      
      while ((j < NO_OF_SMALL_CELLS) && ContFlag)
      { SLevelPointer[0][j]->ReturnLB(x, y); 
        if (ABS(y - p.y) < distance)
        { for (i = 0; i < NO_OF_SMALL_CELLS; i++)
            counter = AddToSpiral(SLevelPointer[i][j]->ReturnTopOfList(), counter);
          j++;
        }
        else
          ContFlag = false;
      }
      break;
    }  

  }

  return counter;
}


void Proximity2L::InitTheClosePoint(MyPoint& p)
// Method initializes the algorithm for the closest point p search
{ 
  DeleteTheNextClosestPointList();

  int PointIndex = -1;
  ReferencePoint = p; // Let us store the point p
  level = 2;

}


void Proximity2L::InitTheClosePoint(double x, double y)
// Method initializes the algorithm for the closest point p search
{ 
  DeleteTheNextClosestPointList();

  int PointIndex = -1;
  ReferencePoint.SetCoordinate(x, y); // Let us store the point p
  level = 2;

}

void Proximity2L::InitTheClosePoint(MyPoint* p)
// Method initializes the algorithm for the closest point p search
{ 
  DeleteTheNextClosestPointList();

  int PointIndex = -1;
  double x, y;
  p->ReturnXY(x, y);
  ReferencePoint.SetCoordinate(x, y);
  ReferencePoint.SetIndex(p->ReturnIndex()); // Let us store the point p
  level = 2;

}

int Proximity2L::DetermineTheClosestPoint(bool& EqualPointsFlag)
// Method finds the closest point to the point p
// Calling the method once again without calling InitTheClosePoint
// returns index of the next closest point etc.
// EqualPointsFlag = true, if a returned point concides with tested one
{ double distance = MAX_DOUBLE;
  level =-2;
  if (level < 1)
    level = 1;
  DCList* TheClosestPoint = NULL;
  EqualPointsFlag = false;

  DCL*** arr;

  // in which cell point ReferencePoint is
  FirstLevelPosition(&ReferencePoint, iX1, jY1);

  if (level == 1) 
  {
    if ( (arr = dclArray[iX1][jY1]->ReturnSecondLevel()) != NULL) 
    { // there is second level
      TheClosestPoint = FindTheClosestPointInCell(dclArray[iX1][jY1]->ReturnTopOfList(), distance);
      dclArray[iX1][jY1]->SecondLevelPosition(&ReferencePoint, iX2, jY2); 
      TheClosestPoint = FindTheClosestPointInCell2Level(
                            dclArray[iX1][jY1]->ReturnSecondLevel(),
                            iX2, jY2, distance);    
    }
    else // only first level exist
      TheClosestPoint = FindTheClosestPointInCell(dclArray[iX1][jY1]->ReturnTopOfList(), distance);
  }

  DCList* temp;
  double TempD = MAX_DOUBLE;
  bool ContinueFlag = true;

  if (Equal(distance, 0))
  { EqualPointsFlag = true;
    DeleteSpiral();  
    return TheClosestPoint->ReturnIndex();
  }

    
  

  
  do { // let us check all cells which are close enough
       // cells are searched in a spiral centered at (iX1, jY1) 
    ContinueFlag = ConstructSpiral(level, ReferencePoint, iX1, jY1, distance);
    if (ContinueFlag)  // spiral is not empty
    { temp = GetDistanceFromSpiral(TempD); 

      if (Equal(TempD, 0)) // points coincide
        goto toEnd;   
      
       // determine the closest point temp in spiral to the ReferencePoint
      if ((temp != NULL) && (TempD < distance)) // Yes, shuch point exist
      { distance = TempD;
        TheClosestPoint = temp;        
      }    
    }
    
    DeleteSpiral(); // delete this spiral
    level++;        // and increase  the distance of a spiral from reference cell

    if (level > BigDiagonal) // emergency jup out
      return NULL;

  } while (ContinueFlag == true);
  
  if (TheClosestPoint != NULL)
    AddAndMarkTheNextClosestPointList(TheClosestPoint->ReturnPoint());

  level--;
  if (TheClosestPoint != NULL)
    return TheClosestPoint->ReturnIndex();
  else
    return -1;

toEnd:
  EqualPointsFlag = true;
  DeleteSpiral();
  TheClosestPoint = temp;  
  AddAndMarkTheNextClosestPointList(TheClosestPoint->ReturnPoint());
 return TheClosestPoint->ReturnIndex();
}




void Proximity2L::CheckIt(MyPoint* p, int n, int found )
// Method serves only as a proof if the 2levelles searching methods works OK.
// it uses brute force method:
// - p: array of points
// - n: number of points inserted up to know - n is the last inserted point
// - foud: the index of the closest point foundt by Proximity2L
// Methods write the error into Shortest.ERR file
{ int i, index;
  double d = MAX_DOUBLE;
  double d1;
  
  if (n == 11)
  {int qq = 0;}

  for (i = 0; i < n; i++)
  {
    d1 = sqrt(Sqr(p[i].x - p[n].x) +
         Sqr(p[i].y - p[n].y));

    if (d1 < d)
    { d = d1;
      index = i;
    }
  }

  
  if (index != found)
  { 
    if (! Equal( 
                 Sqr(p[index].x - p[n].x) + Sqr(p[index].y - p[n].y),
                 Sqr(p[found].x - p[n].x) + Sqr(p[found].y - p[n].y) ) )
    {
      FILE* ff= fopen("Shortest.ERR", "wt");
      fprintf(ff, "napaka: namesto %d je nasel %d \n", found , index);
      fprintf(ff, "to je iteracije %d \n", n);
      double d2 = sqrt(Sqr(p[found].x - p[n].x) +
                    Sqr(p[found].y - p[n].y));

      fprintf(ff, "namesto %lf je nasel %lf \n", d2, d1);
      fclose(ff);
    }
  } 

}



void Proximity2L::CheckNext(MyPoint* p, int n, int found, int next)
{ int i;
  double dMin = Sqr(p[found].x - p[n].x) + Sqr(p[found].y - p[n].y);
  int index = found;
  double d = MAX_DOUBLE;                 
  double d1;
  
  
  for (i = 0; i < n; i++)
  {
    d1 = Sqr(p[i].x - p[n].x) + Sqr(p[i].y - p[n].y);

    if (d1 != dMin)
      if (d1 < d) 
      { d = d1;
      
        if (p[i].ReturnIndex() != p[found].ReturnIndex())
          index = i;
      }
    
  }

  if (index != next)
  { int qqq = 0; // napaka
 
  } 

}



void Proximity2L::Statistic(char* fileName)
{
  FILE* fs = fopen(fileName, "wt");
  fprintf(fs, "Statistik of 2LUPS \n");
  fprintf(fs,"Bounding box: %lf %lf %lf %lf \n", BoxXmin,BoxYmin,BoxXmax,BoxYmax);
  fprintf(fs, "number of cels in X Y directions: %d %d\n", NoOfCellsX1, NoOfCellsY1);
  fprintf(fs, "Theshold %d \n", Threshold);


  int i, j;
  int NoOfSubdivided = 0;
  int NoOfEmpty = 0;

  for (i = 0; i < NoOfCellsX1; i++)
    for (j = 0; j < NoOfCellsY1; j++)
    { if (dclArray[i][j]->ReturnSecondLevel() != NULL)
        NoOfSubdivided++;

      if (dclArray[i][j]->ReturnNoOfPoints() == 0)
        NoOfEmpty++;
    }

  fprintf(fs, "No of subdivided cells %d \n", NoOfSubdivided);
  fprintf(fs, "No of empty cells %d \n", NoOfEmpty);

  fclose(fs);
}