// Proximity2LNew.cpp: implementation of the Proximity2LNew class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyTriangF.h"
#include "Proximity2LNew.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

static USPCellInfo** FirstLevelArray = NULL;
static MyPointT* PointArray = NULL;
static SpiralType* Spirale = NULL;
static SpiralType* Spirale2 = NULL;
static SpiralType* VisitedCells = NULL;

static double BoxXmin = 0; 
static double BoxYmin = 0;
static double BoxXmax = 0;
static double BoxYmax = 0;
static double XSize1 = 0;
static double YSize1 = 0;
static double SizeSmallX = 0;
static double SizeSmallY = 0;
static int UpperSmallCellsLimit = 0;
static int MiddleSmallCellPosition = 0;

static int Threshold = 0;
static int NoOfCellsX1, NoOfCellsY1;
static long BigDiagonal = 0;
static long ANoOfPointsInCell = 0; // avarage number of points in a cell
static double ReferencePointX = 0;
static double ReferencePointY = 0;
static long ReferencePointIndex = 0;
static int ReferenceCellX = 0;
static int ReferenceCellY = 0;
static int ReferenceCellX2 = 0; 
static int ReferenceCellY2 = 0;
static long SpiraleIndex;
static long SpiraleIndex2;
static long VisitedCellIndex = 0;
static NextClosestP *NextListTail, *NextListHead, *NCPPointer;


int InitProximityStructure(MyPointT* p, double ax1, double ay1, double bx1, double by1, int n)
// prepared 2-leveled uniform grid for the use
// first 4 parameters are bounding square coordinates
// n is total number of points
{ if (n == 0)
    return 0;
  else
  { int i, j;
      
    PointArray = p;

    double a = 0.02 * (bx1 - ax1);
    double b = 0.02 * (by1 - ay1);
  
    BoxXmin = ax1 - a; // bounding box is increased a bit
    BoxYmin = ay1 - b;
    BoxXmax = bx1 + a;
    BoxYmax = by1 + b;
      
    double deltaX = BoxXmax - BoxXmin;
    double deltaY = BoxYmax - BoxYmin;
    double ratioXY = deltaX/deltaY;
    
    // determine number of cells at the first level using some heuristic
    NoOfCellsX1 = (int)(0.5 * sqrt(n) * ratioXY  );
    NoOfCellsY1 = (int)(0.5 * sqrt(n) / ratioXY  );

    // too much cells
    if (NoOfCellsX1 > UPPER_ARRAY_LIMIT)
      NoOfCellsX1 = UPPER_ARRAY_LIMIT;
     
    if (NoOfCellsY1 > UPPER_ARRAY_LIMIT)
      NoOfCellsY1 = UPPER_ARRAY_LIMIT;
 
   // not enough cells

    if (NoOfCellsX1 < 1)
      NoOfCellsX1 = 1;
     
    if (NoOfCellsY1 < 1)
      NoOfCellsY1 = 1;

    int bord = 2*(NoOfCellsX1 * NoOfCellsY1);
    Spirale = new SpiralType[bord]; 

    Spirale2 = new SpiralType[bord];

    UpperSmallCellsLimit = NO_OF_SMALL_CELLS - 1;
    MiddleSmallCellPosition = (int)(NO_OF_SMALL_CELLS / 2);

    BigDiagonal = NoOfCellsX1*NoOfCellsY1 + 10; 

    VisitedCells = new SpiralType[NoOfCellsX1*NoOfCellsY1];
    VisitedCellIndex = 0;

    // determine size of cells at the first level
    XSize1 = deltaX / NoOfCellsX1;
    YSize1 = deltaY / NoOfCellsY1;

    // determine number of points in the cells which couses the second level is introduced

   Threshold = (int)(4 * n/(NoOfCellsX1 * NoOfCellsY1) );

// Threshold = 100;

//Threshold = 12;

    // the memory is allocated for the first level
    FirstLevelArray = new USPCellInfo*[NoOfCellsX1];
    for (i = 0; i < NoOfCellsX1; i++)
      FirstLevelArray[i] = new USPCellInfo[NoOfCellsY1];

    
    SizeSmallX = XSize1 / NO_OF_SMALL_CELLS;
    SizeSmallY = YSize1 / NO_OF_SMALL_CELLS;
    double le, bo;

    for (i = 0; i < NoOfCellsX1; i++)
      for (j = 0; j < NoOfCellsY1; j++)
      { le = BoxXmin + i * XSize1;
        bo = BoxYmin + j * YSize1;
        FirstLevelArray[i][j].Left = le;
        FirstLevelArray[i][j].Bottom = bo;
        FirstLevelArray[i][j].Right = le + XSize1;
        FirstLevelArray[i][j].Top = bo + YSize1;
        FirstLevelArray[i][j].NoOfPoints = 0;
        FirstLevelArray[i][j].Points1L = new long[Threshold];
        FirstLevelArray[i][j].SecondLevel = NULL;
        FirstLevelArray[i][j].Points1LChecked = false;
      }
  }

  NextListTail = NextListHead = new NextClosestP;
  NextListTail->next = NULL;
  NCPPointer = NULL;
  return 1;
}

void ReleaseProximityStructure()
{ int i,j, m, n;
  USPPointInfo *p, *list;
  USPCellInfo*** matrix2;

  for (i = 0; i < NoOfCellsX1; i++)
  { for (j = 0; j < NoOfCellsY1; j++)
    { delete[] FirstLevelArray[i][j].Points1L;

      if ((matrix2 = FirstLevelArray[i][j].SecondLevel) != NULL)
      { for (m = 0; m < NO_OF_SMALL_CELLS; m++)
        { for (n = 0; n < NO_OF_SMALL_CELLS; n++)
          { list = matrix2[m][n]->Points2L;
            while (list != NULL)
            { p = list;
              list = list->next;
              delete p;
            }    
            delete matrix2[m][n];       
          }
          delete[] matrix2[m];
        }
        delete[] matrix2;
      }        
    }
    delete[] FirstLevelArray[i]; 
  }  

  delete[] FirstLevelArray;
  FirstLevelArray = NULL;

  if (Spirale != NULL)
  { delete[] Spirale;
    Spirale = NULL;
  }

  if (Spirale2 != NULL)
  { delete[] Spirale2;
    Spirale2 = NULL;
  }

  if (VisitedCells != NULL)
  { delete[] VisitedCells;
    VisitedCells = NULL;
  }

  while (NextListHead != NULL)
  { NextListTail = NextListHead;
    NextListHead = NextListHead->next;
    delete NextListTail;
  }

}



void FirstLevelPosition(long ind, int& PosX, int& PosY)
// method determines the coordinate of the cell into which point p falls
{ 
  PosX = (int)((PointArray[ind].x - BoxXmin)/XSize1);
  PosY = (int)((PointArray[ind].y - BoxYmin)/YSize1);
}



void InsertPoint2USP(long ind)
{ int PosX = (int)((PointArray[ind].x - BoxXmin)/XSize1); 
  int PosY = (int)((PointArray[ind].y - BoxYmin)/YSize1);
 
  int position;
  if ((position=FirstLevelArray[PosX][PosY].NoOfPoints) < Threshold) // we are inserting in the first level
  { FirstLevelArray[PosX][PosY].NoOfPoints++;
    FirstLevelArray[PosX][PosY].Points1L[position] = ind;
  }
  else
  { USPCellInfo*** cell;

    if (FirstLevelArray[PosX][PosY].SecondLevel == NULL)
    { int i, j;
      // CreateSecondLevel 
      // memory is alocated
      FirstLevelArray[PosX][PosY].SecondLevel = new USPCellInfo**[NO_OF_SMALL_CELLS];
      for (i = 0; i < NO_OF_SMALL_CELLS; i++) 
        FirstLevelArray[PosX][PosY].SecondLevel[i] = new USPCellInfo*[NO_OF_SMALL_CELLS];
    
      double l, b;
      // cells are initialised
      cell = FirstLevelArray[PosX][PosY].SecondLevel;
      for (i = 0; i < NO_OF_SMALL_CELLS; i++)
        for (j = 0; j < NO_OF_SMALL_CELLS;j++)
          cell[i][j] = new USPCellInfo;
       

      l = FirstLevelArray[PosX][PosY].Left - SizeSmallX;
      for (i = 0; i < NO_OF_SMALL_CELLS; i++)
      { l += SizeSmallX;
        b = FirstLevelArray[PosX][PosY].Bottom - SizeSmallY;
        for (j = 0; j < NO_OF_SMALL_CELLS; j++)
        { b += SizeSmallY;          
          cell[i][j]->Left = l;
          cell[i][j]->Bottom = b;
          cell[i][j]->Right = l + SizeSmallX;
          cell[i][j]->Top = b + SizeSmallY;
          cell[i][j]->Points2L = NULL;
        }  
      } 
    } 
    else 
      cell = FirstLevelArray[PosX][PosY].SecondLevel;

     // insert the point in the second level now   
    int PosX2 = (int)((PointArray[ind].x - FirstLevelArray[PosX][PosY].Left)/SizeSmallX);
    int PosY2 = (int)((PointArray[ind].y - FirstLevelArray[PosX][PosY].Bottom)/SizeSmallY);

    USPPointInfo* s = new USPPointInfo;
    s->index = ind;
    s->next = cell[PosX2][PosY2]->Points2L;       
    cell[PosX2][PosY2]->Points2L = s;
  }  
}

///////////
void ExaminePointsInCellAtLevel_1(int PosX, int PosY, double& dis, long& CIndex)
{ double a, b, d;
  long pInd;

  if (FirstLevelArray[PosX][PosY].Points1LChecked == false)
  { FirstLevelArray[PosX][PosY].Points1LChecked = true;
    VisitedCells[VisitedCellIndex].CellX1 = PosX;
    VisitedCells[VisitedCellIndex++].CellY1 = PosY;

    for (int i = 0; i < FirstLevelArray[PosX][PosY].NoOfPoints; i++)
    { pInd = FirstLevelArray[PosX][PosY].Points1L[i];
      a = PointArray[pInd].x - ReferencePointX; 
      b = PointArray[pInd].y - ReferencePointY;
      d = (a * a + b * b);
  
      if (Equal(d, 0)) // points are the same
      { CIndex = THE_SAME_POINTS;
        goto konec;
      }

      if ( d < dis )
      { dis = d;
        CIndex = pInd;
        NCPPointer = NextListTail;
      }

      // now we store the point for the NextClosestPoint search
      NextListTail->dist = d;
      NextListTail->index = pInd;
      NextListTail->UsedFlag = false;
      if (NextListTail->next == NULL)
      { NextClosestP *q = new NextClosestP;
        q->next = NULL;
        NextListTail->next = q; 
      }
      NextListTail = NextListTail->next;
    }
  }
konec: ;
}


void ExaminePointsInCellAtLevel_2(USPPointInfo* list, double& dis, long& CIndex)
{ long pInd;
  double a, b, d;

  if (list != NULL)
  { if (list->Points2LChecked == false)
    { list->Points2LChecked = true;
      while (list != NULL)
      { pInd = list->index;
        a = PointArray[pInd].x - ReferencePointX; 
        b = PointArray[pInd].y - ReferencePointY;
        d = a * a + b * b; 

        if (Equal(d, 0))
        { CIndex = THE_SAME_POINTS;
          goto konec;
        }

        if ( d < dis )
        { dis = d;
          CIndex = pInd;
          NCPPointer = NextListTail;
        }

        // now we store the point for the NextClosestPoint search
        NextListTail->dist = d;
        NextListTail->index = pInd;
        NextListTail->UsedFlag = false;
        if (NextListTail->next == NULL)
        { NextClosestP *q = new NextClosestP;
          q->next = NULL;
          NextListTail->next = q; 
        }
        NextListTail = NextListTail->next;
        list = list->next;
      }
    }
  konec: ;
  }
}


void AddUpperCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ 
  int psk = PosY1 + 1;
  if (psk < NoOfCellsY1) // one cell up
  { double a = FirstLevelArray[PosX1][PosY1].Top - ReferencePointY;
    if (dis > a * a)
    { 

      int PosX2 = (int)((PointArray[ReferencePointIndex].x - FirstLevelArray[PosX1][PosY1].Left)/SizeSmallX);
      if (PosX2 < 0) PosX2 = 0;
      if (PosX2 > NO_OF_SMALL_CELLS-1) PosX2 = NO_OF_SMALL_CELLS-1;
//    int PosY2 = (int)((PointArray[ind].y - FirstLevelArray[PosX][PosY].Bottom)/SizeSmallY);

      TheClosestPointInSecondLevel(PosX1, psk, PosX2, 0, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = PosX1;
      Spirale[SpiraleIndex].CellY1 = psk;
      Spirale[SpiraleIndex++].code = UPPER_CELL;
      ExaminePointsInCellAtLevel_1(PosX1, psk, dis, ClosestI);
    }
  }   
}


void AddLowerCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{
  int psk = PosY1 - 1;
   if (psk >= 0) 
  { double a = ReferencePointY - FirstLevelArray[PosX1][PosY1].Bottom;
    if (dis > a * a)
    { 

      int PosX2 = (int)((PointArray[ReferencePointIndex].x - FirstLevelArray[PosX1][PosY1].Left)/SizeSmallX);
      if (PosX2 < 0) PosX2 = 0;
      if (PosX2 > NO_OF_SMALL_CELLS-1) PosX2 = NO_OF_SMALL_CELLS-1;
//    int PosY2 = (int)((PointArray[ind].y - FirstLevelArray[PosX][PosY].Bottom)/SizeSmallY);

      TheClosestPointInSecondLevel(PosX1, psk, PosX2, UpperSmallCellsLimit, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = PosX1;
      Spirale[SpiraleIndex].CellY1 = psk;
      Spirale[SpiraleIndex++].code = LOWER_CELL;
      ExaminePointsInCellAtLevel_1(PosX1, psk, dis, ClosestI);
    }
  }   
}


void AddLeftCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ 
  int psk = PosX1 - 1;
  if (psk >= 0) 
  { double a = ReferencePointX - FirstLevelArray[PosX1][PosY1].Left;
    if (dis > a * a)
    { 

//    int PosX2 = (int)((PointArray[ind].x - FirstLevelArray[PosX][PosY].Left)/SizeSmallX);
      int PosY2 = (int)((PointArray[ReferencePointIndex].y - FirstLevelArray[PosX1][PosY1].Bottom)/SizeSmallY);
      if (PosY2 < 0) PosY2 = 0;
      if (PosY2 > NO_OF_SMALL_CELLS-1) PosY2 = NO_OF_SMALL_CELLS-1;

      TheClosestPointInSecondLevel(psk, PosY1, UpperSmallCellsLimit, PosY2, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = psk;
      Spirale[SpiraleIndex].CellY1 = PosY1;
      Spirale[SpiraleIndex++].code = LEFT_CELL;
      ExaminePointsInCellAtLevel_1(psk, PosY1, dis, ClosestI);
    }
  }   
}



void AddRightCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ int psk = PosX1 + 1;
  if (psk < NoOfCellsX1) 
  { double a = FirstLevelArray[PosX1][PosY1].Right - ReferencePointX;
    if (dis > a * a)
    {

//    int PosX2 = (int)((PointArray[ReferencePointIndex].x - FirstLevelArray[PosX1][PosY1].Left)/SizeSmallX);
      int PosY2 = (int)((PointArray[ReferencePointIndex].y - FirstLevelArray[PosX1][PosY1].Bottom)/SizeSmallY);
      if (PosY2 < 0) PosY2 = 0;
      if (PosY2 > NO_OF_SMALL_CELLS-1) PosY2 = NO_OF_SMALL_CELLS-1;

      TheClosestPointInSecondLevel(psk, PosY1, 0, PosY2, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = psk;
      Spirale[SpiraleIndex].CellY1 = PosY1;
      Spirale[SpiraleIndex++].code = RIGHT_CELL;
      ExaminePointsInCellAtLevel_1(psk, PosY1, dis, ClosestI);
    }
  }   
}


void AddUpperLeftCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ if ((PosX1 != ReferenceCellX) && (PosY1 != ReferenceCellY)) // add three cells
  { AddLeftCell(PosX1, PosY1, dis, ClosestI);
    AddUpperCell(PosX1, PosY1, dis, ClosestI);
  }

  int pskx = PosX1 - 1;
  int psky = PosY1 + 1;
  if ((pskx >= 0) && (psky < NoOfCellsY1))
  { double a = FirstLevelArray[PosX1][PosY1].Left - ReferencePointX;
    double b = ReferencePointY - FirstLevelArray[PosX1][PosY1].Top;
    if (dis > a * a + b * b)
    { 
      TheClosestPointInSecondLevel(pskx, psky, UpperSmallCellsLimit, 0, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = pskx;
      Spirale[SpiraleIndex].CellY1 = psky;
      Spirale[SpiraleIndex++].code = LEFT_UPPER_CELL;
      ExaminePointsInCellAtLevel_1(pskx, psky, dis, ClosestI);
    }
  }
}


void AddUpperRightCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ if ((PosX1 != ReferenceCellX) && (PosY1 != ReferenceCellY)) // add three cells
  { AddRightCell(PosX1, PosY1, dis, ClosestI);
    AddUpperCell(PosX1, PosY1, dis, ClosestI);
  }

  int pskx = PosX1 + 1;
  int psky = PosY1 + 1;
  if ((pskx < NoOfCellsX1) && (psky < NoOfCellsY1))
  { double a = FirstLevelArray[PosX1][PosY1].Right - ReferencePointX;
    double b = ReferencePointY - FirstLevelArray[PosX1][PosY1].Top;
    if (dis > a * a + b * b)
    { 
      TheClosestPointInSecondLevel(pskx, psky, 0, 0, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = pskx;
      Spirale[SpiraleIndex].CellY1 = psky;
      Spirale[SpiraleIndex++].code = RIGHT_UPPER_CELL;
      ExaminePointsInCellAtLevel_1(pskx, psky, dis, ClosestI);
    }
  }
}



void AddLowerLeftCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ if ((PosX1 != ReferenceCellX) && (PosY1 != ReferenceCellY)) // add three cells
  { AddLeftCell(PosX1, PosY1, dis, ClosestI);
    AddLowerCell(PosX1, PosY1, dis, ClosestI);
  }

  int pskx = PosX1 - 1;
  int psky = PosY1 - 1;
  if ((pskx >= 0) && (psky >= 0))
  { double a = FirstLevelArray[PosX1][PosY1].Left - ReferencePointX;
    double b = ReferencePointY - FirstLevelArray[PosX1][PosY1].Bottom;
    if (dis > a * a + b * b)
    { 
     TheClosestPointInSecondLevel(pskx, psky, UpperSmallCellsLimit, UpperSmallCellsLimit, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = pskx;
      Spirale[SpiraleIndex].CellY1 = psky;
      Spirale[SpiraleIndex++].code = LEFT_LOWER_CELL;
      ExaminePointsInCellAtLevel_1(pskx, psky, dis, ClosestI);
    }
  }
}


void AddLowerRightCell(int PosX1, int PosY1, double& dis, long& ClosestI)
{ if ((PosX1 != ReferenceCellX) && (PosY1 != ReferenceCellY)) // add three cells
  { AddRightCell(PosX1, PosY1, dis, ClosestI);
    AddLowerCell(PosX1, PosY1, dis, ClosestI);
  }

  int pskx = PosX1 + 1;
  int psky = PosY1 - 1;
  if ((pskx < NoOfCellsX1) && (psky >= 0))
  { double a = FirstLevelArray[PosX1][PosY1].Right - ReferencePointX;
    double b = ReferencePointY - FirstLevelArray[PosX1][PosY1].Bottom;
    if (dis > a * a + b * b)
    { 
      TheClosestPointInSecondLevel(pskx, psky, 0, UpperSmallCellsLimit, dis, ClosestI);

      Spirale[SpiraleIndex].CellX1 = pskx;
      Spirale[SpiraleIndex].CellY1 = psky;
      Spirale[SpiraleIndex++].code = RIGHT_LOWER_CELL;
      ExaminePointsInCellAtLevel_1(pskx, psky, dis, ClosestI);
    }
  }
}

void AddNeighboursToTheSpiral_1_Level(int PosX1, int PosY1, double& dis, long &ClosestI, int code)
{ switch (code)
  { case MIDDLE_CELL:
    { AddUpperCell(PosX1, PosY1, dis, ClosestI);
      AddLowerCell(PosX1, PosY1, dis, ClosestI);
      AddLeftCell(PosX1, PosY1, dis, ClosestI);
      AddRightCell(PosX1, PosY1, dis, ClosestI);
      AddUpperLeftCell(PosX1, PosY1, dis, ClosestI);
      AddUpperRightCell(PosX1, PosY1, dis, ClosestI);
      AddLowerLeftCell(PosX1, PosY1, dis, ClosestI);
      AddLowerRightCell(PosX1, PosY1, dis, ClosestI);
      break;
    }
    case UPPER_CELL: AddUpperCell(PosX1, PosY1, dis, ClosestI); break;
    case LOWER_CELL: AddLowerCell(PosX1, PosY1, dis, ClosestI); break;
    case LEFT_CELL:  AddLeftCell(PosX1, PosY1, dis, ClosestI); break;
    case RIGHT_CELL: AddRightCell(PosX1, PosY1, dis, ClosestI); break;
    case LEFT_UPPER_CELL: AddUpperLeftCell(PosX1, PosY1, dis, ClosestI); break;
    case RIGHT_UPPER_CELL: AddUpperRightCell(PosX1, PosY1, dis, ClosestI); break;
    case LEFT_LOWER_CELL: AddLowerLeftCell(PosX1, PosY1, dis, ClosestI); break;
    case RIGHT_LOWER_CELL: AddLowerRightCell(PosX1, PosY1, dis, ClosestI); break;
  }
}

/////////////level 2

void AddUpperCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if (PosY + 1 < NO_OF_SMALL_CELLS) // one cell up
  { double a = level2[PosX][PosY]->Top - ReferencePointY;
    if (dis > a * a)
    { Spirale2[SpiraleIndex2].CellX1 = PosX;
      Spirale2[SpiraleIndex2].CellY1 = PosY + 1;
      Spirale2[SpiraleIndex2++].code = UPPER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX][PosY+1]->Points2L, dis, ClosestI);

    }
  }   
}


void AddLowerCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if (PosY - 1 >= 0) 
  { double a = ReferencePointY - level2[PosX][PosY]->Bottom;
    if (dis > a * a)
    { Spirale2[SpiraleIndex2].CellX1 = PosX;
      Spirale2[SpiraleIndex2].CellY1 = PosY - 1;
      Spirale2[SpiraleIndex2++].code = LOWER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX][PosY - 1]->Points2L, dis, ClosestI);
    }
  }   
}


void AddLeftCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if (PosX - 1 >= 0) 
  { double a = ReferencePointX - level2[PosX][PosY]->Left;
    if (dis > a * a)
    { Spirale2[SpiraleIndex2].CellX1 = PosX - 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY;
      Spirale2[SpiraleIndex2++].code = LEFT_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX - 1][PosY]->Points2L, dis, ClosestI);
    }
  }   
}


void AddRightCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if (PosX + 1 < NO_OF_SMALL_CELLS) 
  { double a = level2[PosX][PosY]->Right - ReferencePointX;
    if (dis >  a * a)
    { Spirale2[SpiraleIndex2].CellX1 = PosX + 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY;
      Spirale2[SpiraleIndex2++].code = RIGHT_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX + 1][PosY]->Points2L, dis, ClosestI);
    }
  }   
}



void AddUpperLeftCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if ((PosX != ReferenceCellX2) && (PosY != ReferenceCellY2)) // add three cells
  { AddLeftCell2(level2, PosX, PosY, dis, ClosestI);
    AddUpperCell2(level2, PosX, PosY, dis, ClosestI);
  }

  if ((PosX - 1 >= 0) && (PosY + 1 < NO_OF_SMALL_CELLS))
  { double a = level2[PosX][PosY]->Left - ReferencePointX;
    double b = ReferencePointY - level2[PosX][PosY]->Top;
    if (dis > a * a + b * b)
    { Spirale2[SpiraleIndex2].CellX1 = PosX - 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY + 1;
      Spirale2[SpiraleIndex2++].code = LEFT_UPPER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX - 1][PosY + 1]->Points2L, dis, ClosestI);
    }
  }
}


void AddUpperRightCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if ((PosX != ReferenceCellX2) && (PosY != ReferenceCellY2)) // add three cells
  { AddRightCell2(level2, PosX, PosY, dis, ClosestI);
    AddUpperCell2(level2, PosX, PosY, dis, ClosestI);
  }

  if ((PosX + 1 < NO_OF_SMALL_CELLS) && (PosY + 1 < NO_OF_SMALL_CELLS))
  { double a = level2[PosX][PosY]->Right - ReferencePointX;
    double b = ReferencePointY - level2[PosX][PosY]->Top;
    if (dis > a * a + b * b )
    { Spirale2[SpiraleIndex2].CellX1 = PosX + 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY + 1;
      Spirale2[SpiraleIndex2++].code = RIGHT_UPPER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX + 1][PosY + 1]->Points2L, dis, ClosestI);
    }
  }
}



void AddLowerLeftCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if ((PosX != ReferenceCellX2) && (PosY != ReferenceCellY2)) // add three cells
  { AddLeftCell2(level2, PosX, PosY, dis, ClosestI);
    AddLowerCell2(level2, PosX, PosY, dis, ClosestI);
  }

  if ((PosX - 1 >= 0) && (PosY - 1 >= 0))
  { double a = level2[PosX][PosY]->Left - ReferencePointX;
    double b = ReferencePointY - level2[PosX][PosY]->Bottom;
    if (dis > a * a + b * b )
    { Spirale2[SpiraleIndex2].CellX1 = PosX - 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY - 1;
      Spirale2[SpiraleIndex2++].code = LEFT_LOWER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX - 1][PosY - 1]->Points2L, dis, ClosestI);
    }
  }
}


void AddLowerRightCell2(USPCellInfo*** level2, int PosX, int PosY, double& dis, long& ClosestI)
{ if ((PosX != ReferenceCellX2) && (PosY != ReferenceCellY2)) // add three cells
  { AddRightCell2(level2, PosX, PosY, dis, ClosestI);
    AddLowerCell2(level2, PosX, PosY, dis, ClosestI);
  }

  if ((PosX + 1 < NO_OF_SMALL_CELLS) && (PosY - 1 >= 0))
  { double a = level2[PosX][PosY]->Right - ReferencePointX;
    double b = ReferencePointY - level2[PosX][PosY]->Bottom;
    if (dis > a * a + b * b )
    { Spirale2[SpiraleIndex2].CellX1 = PosX + 1;
      Spirale2[SpiraleIndex2].CellY1 = PosY - 1;
      Spirale2[SpiraleIndex2++].code = RIGHT_LOWER_CELL;
      ExaminePointsInCellAtLevel_2(level2[PosX + 1][PosY - 1]->Points2L, dis, ClosestI);
    }
  }
}


void AddNeighboursToTheSpiral_2_Level(USPCellInfo*** level2, int PosX2, int PosY2, double& dis, long &ClosestI, int code)
{ switch (code)
  { case MIDDLE_CELL:
    { AddUpperCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddLowerCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddLeftCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddRightCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddUpperLeftCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddUpperRightCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddLowerLeftCell2(level2, PosX2, PosY2, dis, ClosestI);
      AddLowerRightCell2(level2, PosX2, PosY2, dis, ClosestI);
      break;
    }
    case UPPER_CELL: AddUpperCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case LOWER_CELL: AddLowerCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case LEFT_CELL:  AddLeftCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case RIGHT_CELL: AddRightCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case LEFT_UPPER_CELL: AddUpperLeftCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case RIGHT_UPPER_CELL: AddUpperRightCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case LEFT_LOWER_CELL: AddLowerLeftCell2(level2, PosX2, PosY2, dis, ClosestI); break;
    case RIGHT_LOWER_CELL: AddLowerRightCell2(level2, PosX2, PosY2, dis, ClosestI); break;
  }
}


void  TheClosestPointInSecondLevel(int PosX, int PosY, int PosX2, int PosY2, double& dis, long& TheCIndex)
{ SpiraleIndex2 = 0;
  USPCellInfo*** level2 = FirstLevelArray[PosX][PosY].SecondLevel;

  if (level2 != NULL)
  { ExaminePointsInCellAtLevel_2(level2[PosX2][PosY2]->Points2L, dis, TheCIndex);

    AddNeighboursToTheSpiral_2_Level(level2, PosX2, PosY2, dis, TheCIndex, MIDDLE_CELL);

    long i=0;
    int codeOut;
    long px, py;
    while (i < SpiraleIndex2)
    { px = Spirale2[i].CellX1;
      py = Spirale2[i].CellY1;
      codeOut = Spirale2[i].code;
      AddNeighboursToTheSpiral_2_Level(level2, px, py, dis,  TheCIndex, codeOut);
      i++;
    }
  }

}

//////////end level2


void FindTheClosestPointInCell(int PosX, int PosY, long& TheCIndex, double& dis, int code)
{
  SpiraleIndex = 0;

   // check now the second level if it exist
  if (FirstLevelArray[PosX][PosY].SecondLevel != NULL) 
  { ReferenceCellX2 = (int)((ReferencePointX - FirstLevelArray[PosX][PosY].Left)/SizeSmallX);
    ReferenceCellY2 = (int)((ReferencePointY - FirstLevelArray[PosX][PosY].Bottom)/SizeSmallY);

    TheClosestPointInSecondLevel(PosX, PosY, ReferenceCellX2, ReferenceCellY2, dis, TheCIndex);
  }

  ExaminePointsInCellAtLevel_1(PosX, PosY, dis, TheCIndex);

 
  AddNeighboursToTheSpiral_1_Level(PosX, PosY, dis,  TheCIndex, code);
  

  long i=0;
  int codeOut;
  long px, py;
  while (i < SpiraleIndex)
  { px = Spirale[i].CellX1;
    py = Spirale[i].CellY1;
    codeOut = Spirale[i].code;
    AddNeighboursToTheSpiral_1_Level(px, py, dis,  TheCIndex, codeOut);
    i++;
  }
}


void InitTheClosestPointSearch()
{ int i, k1, k2;
  int PosX, PosY;

  for (i = 0; i < VisitedCellIndex; i++)
  {
    PosX = VisitedCells[i].CellX1;
    PosY = VisitedCells[i].CellY1;

    FirstLevelArray[PosX][PosY].Points1LChecked = false;
    
    if (FirstLevelArray[PosX][PosY].SecondLevel != NULL)
      for (k1 = 0; k1 < NO_OF_SMALL_CELLS; k1++)
        for (k2 = 0; k2 < NO_OF_SMALL_CELLS; k2++)
        { if (FirstLevelArray[PosX][PosY].SecondLevel[k1][k2]->Points2L != NULL)
            FirstLevelArray[PosX][PosY].SecondLevel[k1][k2]->Points2L->Points2LChecked = false;

        }
      
  }
  VisitedCellIndex = 0;
}


long DetermineTheClosestPoint(long refPI)
{ 
  
  InitTheClosestPointSearch();

  ReferencePointX = PointArray[refPI].x;
  ReferencePointY = PointArray[refPI].y;
  ReferencePointIndex = refPI;

  ReferenceCellX = (int)((ReferencePointX - BoxXmin)/XSize1); 
  ReferenceCellY = (int)((ReferencePointY - BoxYmin)/YSize1);

  long TheClosestIndex = NONE;
  double CurrentDistance = MAX_DOUBLE;
  NextListTail = NextListHead;
  NCPPointer = NULL;

  FindTheClosestPointInCell(ReferenceCellX, ReferenceCellY, TheClosestIndex, CurrentDistance, MIDDLE_CELL);

  if (NCPPointer != NULL)
    NCPPointer->UsedFlag = true;  // mark the Closest point

  return TheClosestIndex;
}

long ReturnTheNextClosestPoint()
{

  // at first we check the points already found and checked

  long TheClosestIndex = NONE;
  double CurrentDistance = MAX_DOUBLE;
  double d; 
  NextClosestP* listD = NextListHead;

  while ((listD != NextListTail) && (listD->next != NULL)) 
  { if (listD->UsedFlag == false)
    { if ((d = listD->dist) < CurrentDistance)
      { CurrentDistance = d;
        TheClosestIndex = listD->index;
        NCPPointer = listD;
      }
    }
    listD = listD->next;
  }

  FindTheClosestPointInCell(ReferenceCellX, ReferenceCellY, TheClosestIndex, CurrentDistance, MIDDLE_CELL);

  
  NCPPointer->UsedFlag = true;  // mark the Closest point

  return TheClosestIndex;
}


