// 
// Finding the closest point using 2 leveled subdivision

// University of Maribor
// Faculty of Electrical Engineering & Computer Science
// Laboratory for Geometric Modelling & Multimedia Algorithms - GEMMA
// Written By Borut ZALIK in August 2000
// e-mail: zalik@uni-mb.si
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PROXIMITY2L_H__40A28027_6B0D_11D4_99BB_004F4902DA76__INCLUDED_)
#define AFX_PROXIMITY2L_H__40A28027_6B0D_11D4_99BB_004F4902DA76__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "MyPoint.h"


#define UPPER_ARRAY_LIMIT 150
//#define NO_OF_SMALL_CELLS 3


class Proximity2L;


class DCList  // list of points
{ MyPoint* p;

public:
  DCList *Next;

  DCList(MyPoint* a)
  { p = a;
    Next = NULL;
  }
   
  ~DCList() {;}

  MyPoint* ReturnPoint() {return p;}
  int ReturnIndex() {return p->ReturnIndex();}
  void ReturnPointCoordinate(double& x, double& y)
  { p->ReturnXY(x, y);}


};


class DCL
// the first level cell
{
private:
  DCList *top;                      // pointer to the list of points 
                                    // at the first level; it is NULL if second level exist
  DCL*** SecondLevelArray;          // second level array
  int NoOfPoints;                   // No of points in the cell
  double Left, Bottom;              // left,bottom corner of the cell
  double Right, Upper;              // right, upper corner of the cell
  double SizeSmallX, SizeSmallY;    // size of small cell 

  // private methods
  void AddPoint(MyPoint* a);
  void AddPoint(DCList* temp);
  void DeleteList();
  bool VisitedFlag;

public:
  DCL(double l, double b, double r, double u, double sx, double sy);
  ~DCL();

  void PlotGrid(HDC hdc, double a1, double b1,
                            double a2, double b2,
                            int rasterX, int rasterY);

  DCL*** ReturnSecondLevel() {return SecondLevelArray;}
  void InsertPoint(MyPoint* p);
  void SecondLevelPosition(MyPoint* p, int& PosX, int& PosY);
  bool ReturnVisitedFlag() {return VisitedFlag;}
  void ResetVisitedFlag() {VisitedFlag = false;}
  DCList* ReturnTopOfList() {return top;}
  void ReturnRU(double& x, double& y) {x = Right; y = Upper;}
  void ReturnRB(double& x, double& y) {x = Right; y = Bottom;}
  void ReturnLU(double& x, double& y) {x = Left; y = Upper;}
  void ReturnLB(double& x, double& y) {x = Left; y = Bottom;}
  int ReturnNoOfPoints() {return NoOfPoints;}
};


class SpiralClass
// one way connected list
{
private:
  DCList* ListPointer;   // pointer to the list of points storred at a cells
  
public:
  SpiralClass* Next;      

  SpiralClass() {;}
  SpiralClass(DCList* l)
  { ListPointer = l;}
  
  ~SpiralClass() {;}

  DCList* ReturnList() {return ListPointer;}
};


class Proximity2L  // Proximity class using Two level uniform subdivision
{
private:

  DCL*** dclArray;        // first level array of double connected lists
  int iX1, jY1,           // coordinate of a point at the first level 
      iX2, jY2;           // coordinate of a point at the second level 
  MyPoint ReferencePoint; // point to which the closest point is searched
  double BoxXmin, BoxYmin, 
         BoxXmax,BoxYmax;  // Bounding box of the region with points
  double XSize1, YSize1;   // size of cells at the first level
  int NoOfCellsX1, NoOfCellsY1; // no. of cells at the first level
  int BigDiagonal;         // number of cells in the diagonal at the first level
  SpiralClass *spiral;     // pointer to the cells surrounding reference point
  int level;               // level of the spiral;
  DCList* TheNextClosestPointList; // list of the points being visited while searching for the
                                   // next closest point

  DCList* AuxDCLList;      // used only if the points are inserted in the form (x, y, ind)
  // private methods

  void DeleteSpiral();
  void DeleteSecondLevel(DCL*** SLA);
  void FirstLevelPosition(MyPoint* p, int& PosX, int& PosY);
  int FindTheClosestPoint();
  DCList*  GetDistanceFromSpiral(double& CurrentDistance);
  DCList* FindTheClosestPointInCell(DCList* list, double& d);
  bool ConstructSpiral(int level, MyPoint p, int i1, int j1, double distance);
  int AddSecondLevelCells(int direction, DCL*** SLevelPointer, MyPoint& p, 
                          double distance, int& counter);
  int AddToSpiral(DCList* l, int counter);
  DCList* FindTheClosestPointInCell2Level( DCL*** Array2,
                          int ii, int jj, double& distance);

  // managing of the list of points which are "the next closest points"
  void AddAndMarkTheNextClosestPointList(MyPoint* q);
  void DeleteTheNextClosestPointList();
public:
	Proximity2L(double ax1, double ay1, double bx1, double by1, int n);
	~Proximity2L();

  void PlotGrid(HDC hdc, int right, int bottom);
  void InsertPoint(MyPoint* p);
  void InsertPoint(double xx, double yy, long ind);
  void InitTheClosePoint(MyPoint& p);
  void InitTheClosePoint(double x, double y);
  void InitTheClosePoint(MyPoint* p);
  int DetermineTheClosestPoint(bool& EqualPointsFlag);


  void CheckIt(MyPoint* , int n, int found ); // samo za test
  void CheckNext(MyPoint* p, int n, int found, int next);
  void Statistic(char* fileName);
};

#endif // !defined(AFX_TLUPDDL_H__40A28027_6B0D_11D4_99BB_004F4902DA76__INCLUDED_)
