// SimpleSkipList.cpp: implementation of the SimpleSkipList class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyTriangF.h"
#include "SimpleSkipList.h"

#include <stdio.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#include "globals.h"
#include "utility.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DoubleLinkedList* DCLLeft;
DoubleLinkedList* DCLRight;


SimpleSkipList::SimpleSkipList(long np, long fraction, double min, double max)
{
fraction = 100;
  NoOfCells = (long)(np / fraction) + 1;
  XMax = max; XMin = min;
  Size = (XMax - XMin)/NoOfCells;

  SkipTable = new DoubleLinkedList*[NoOfCells];

  Left = new DoubleLinkedList;
  Left->EntryIndex = 0;
  Left->key = -MAX_DOUBLE;
  Left->y = 0;
  Left->LPointer = NULL;
  Left->TriangleIndex = -1;
  Left->VertexNo = NOINDEX;
  Left->Next = NULL;

  Right = new DoubleLinkedList;
  Right->EntryIndex = NOINDEX;
  Right->key = MAX_DOUBLE;
  Right->y = 0;
  Right->LPointer = Left;
  Left->RPointer = Right;
  Right->RPointer = NULL;
  Right->TriangleIndex = -1;
  Right->VertexNo = NOINDEX;
  Right->Next = NULL;

  Pool = NULL;

  for (long i = 1; i < NoOfCells; i++) 
    SkipTable[i] = NULL;

  SkipTable[0] = Left;

}

SimpleSkipList::~SimpleSkipList()
{
  delete[] SkipTable;
  SkipTable = NULL;

  DoubleLinkedList* w;
  while (Left != NULL)
  { w = Left;
    Left = Left->RPointer;
    delete w;
  }

  while (Pool != NULL)
  { w = Pool;
    Pool = Pool->Next;
    delete w;    
  }
  Pool = NULL;
}


DoubleLinkedList* SimpleSkipList::ReturnFromPool()
{ 
  if (Pool == NULL)
  { DoubleLinkedList* w =  new DoubleLinkedList;
    w->Next = NULL;
    return w;
  }
   
  DoubleLinkedList* w = Pool;
  Pool = Pool->Next;
  return w;
}

void SimpleSkipList::InsertIntoPool(DoubleLinkedList* w)
{ if (w != NULL)
  { w->Next = Pool;
    Pool = w;
  }
}

void SimpleSkipList::AddAfter(DoubleLinkedList* OldLeft, DoubleLinkedList* NewE)
{ DoubleLinkedList* OldRight = OldLeft->RPointer;

  OldLeft->RPointer = NewE;
  NewE->LPointer = OldLeft;
  NewE->RPointer = OldRight;
  OldRight->LPointer = NewE;
  NewE->LPointer->TriangleIndex = NewE->TriangleIndex;
 
}


void SimpleSkipList::AddBefore(DoubleLinkedList* OldRight, DoubleLinkedList* NewE)
{ DoubleLinkedList* OldLeft = OldRight->LPointer;

  OldLeft->RPointer = NewE;
  NewE->LPointer = OldLeft;
  OldRight->LPointer = NewE;
  NewE->RPointer = OldRight;
  NewE->LPointer->TriangleIndex = NewE->TriangleIndex;
 
}

int SimpleSkipList::AppendAtTheLeft(double x, double y, long v, long TIndex)
{ DoubleLinkedList* a = Left->RPointer;
  if (TheSamePoints2D(a->key, a->y, x, y))
    return 0;

//  DoubleLinkedList* NewEl = new DoubleLinkedList;

  DoubleLinkedList* NewEl = ReturnFromPool();

  NewEl->EntryIndex = 0;
  NewEl->key = x;
  NewEl->VertexNo = v;
  NewEl->y = y;
  NewEl->TriangleIndex = TIndex;
  

  // the care is take when inserted point has the same 
  // x coordinate as the first point(s)
  // in this case, the new record is assigned after
  // the first point
  // added on May 30. 2003

  if (Equal(x, a->key))
  { while (Equal(x, a->key))
      a = a->RPointer;

    a->LPointer->TriangleIndex = TIndex;
    NewEl->LPointer = a->LPointer;
    NewEl->RPointer = a;
    a->LPointer->RPointer = NewEl;
    a->LPointer = NewEl;
    return 1;
  }

  // end of addtitions



  Left->RPointer = NewEl;
  NewEl->RPointer = a;
  a->LPointer = NewEl;
  NewEl->LPointer = Left;


  long EntryIndex = (long)((x - XMin)/Size);
  if (EntryIndex >= NoOfCells)
    EntryIndex = NoOfCells-1;
 
  if (EntryIndex == 0)
    return 1;

  if (SkipTable[EntryIndex] == NULL)
  { SkipTable[EntryIndex] = NewEl;
    NewEl->EntryIndex = EntryIndex;
    return 1;
  }

  if (x < SkipTable[EntryIndex]->key)
  { SkipTable[EntryIndex] = NewEl;
    NewEl->EntryIndex = EntryIndex;
    return 1;
  }

  NewEl->EntryIndex = EntryIndex;
  return 1;
}

int SimpleSkipList::AppendAtTheRight(double x, double y, long v, long TIndex)
{
  return Insert(x,y,v, TIndex);

}

int SimpleSkipList::Insert(double x, double y, long vertex, long TIndex)
{ long EntryIndex = (long)((x - XMin)/Size);
  if (EntryIndex >= NoOfCells)
    EntryIndex = NoOfCells-1;
 
//  DoubleLinkedList* NewEl = new DoubleLinkedList;
  DoubleLinkedList* NewEl = ReturnFromPool();

  NewEl->EntryIndex = EntryIndex;
  NewEl->key = x;
  NewEl->VertexNo = vertex;
  NewEl->y = y;
  NewEl->TriangleIndex = TIndex;

  DoubleLinkedList* walker;

  if (SkipTable[EntryIndex] == NULL)
  { // at this location there is no entries yet;
    long i = EntryIndex -1;
    while (SkipTable[i] == NULL) i--;
 
    walker = SkipTable[i];
    while (walker->RPointer->key < x) 
      walker = walker->RPointer;
    
    AddAfter(walker, NewEl);
    SkipTable[EntryIndex] = NewEl;
    return 1;
  }

  
  walker = SkipTable[EntryIndex];
  while (walker->RPointer->key < x) 
    walker = walker->RPointer;

  if (TheSamePoints2D(walker->key, walker->y, x, y))
  { // delete NewEl;
InsertIntoPool(NewEl);
    return 0;
  }

  if (walker == SkipTable[EntryIndex])
  {
    if (walker->key < x) 
      AddAfter(walker, NewEl);
    else
    { AddBefore(walker, NewEl);
      SkipTable[EntryIndex] = NewEl;
    }
  }
  else
    AddAfter(walker, NewEl);
  return 1;
}






int SimpleSkipList::ReturnInterval(double x, double y, 
                                   DoubleLinkedList** LeftOut, 
                                   DoubleLinkedList** RightOut)
{ long EntryIndex = (long)((x - XMin)/Size);
  if (EntryIndex >= NoOfCells)
    EntryIndex = NoOfCells-1;
  
  while (SkipTable[EntryIndex] == NULL)
    EntryIndex--;

  DoubleLinkedList* w = SkipTable[EntryIndex];

  while (w->key < x) w = w->RPointer;

  if (w == Right)
  { *RightOut = CurrentRight = NULL;
    *LeftOut = CurrentLeft = w->LPointer;
     return RIGHT;
  }

  if (w->LPointer == Left)
  { 
/*    if (Equal(w->key, x))
    { *LeftOut = CurrentLeft = w;
      *RightOut = CurrentRight = w->RPointer;
      return MIDDLE;
    }
*/
    *RightOut = CurrentRight = w;
    *LeftOut = CurrentLeft = NULL;
     return LEFT;
  } 

  *RightOut = CurrentRight = w;
  *LeftOut = CurrentLeft = w->LPointer;

  if (TheSamePoints2D(x, y,CurrentRight->key, CurrentRight->y))
    return POINTS_THE_SAME; 

  if (Equal(y,CurrentRight->y) && Equal(y, CurrentLeft->y))
    return ON_EDGE;

   return MIDDLE;
}


void SimpleSkipList::ChangeTriangleIndex(long vertex,long TIndex)
{ DoubleLinkedList* w = Left;

  while ((w != NULL) && (w->TriangleIndex != TIndex))
    w = w->RPointer;

  if (w != NULL)
    w->TriangleIndex = TIndex;
}

int SimpleSkipList::UpdateTIndex(double x, long i, long j, long ti)
{ long EntryIndex = (long)((x - XMin)/Size);
  if (EntryIndex >= NoOfCells)
    EntryIndex = NoOfCells-1;
  
  while (SkipTable[EntryIndex] == NULL)
    EntryIndex--;

  DoubleLinkedList* w = SkipTable[EntryIndex];

    
  while (1)
  { if ((w->VertexNo == i) && (w->RPointer->VertexNo == j))
    { w->TriangleIndex = ti;
      return 1;
    }
    if ((w->VertexNo == j) && (w->RPointer->VertexNo == i))
    { w->TriangleIndex = ti;
      return 1;
    }

    if ((w->VertexNo == i) && (w->LPointer->VertexNo == j))
    { w->TriangleIndex = ti;
      return 1;
    }

    if ((w->VertexNo == j) && (w->LPointer->VertexNo == i))
    { w->TriangleIndex = ti;
      return 1;
    }

    if (w->key > x)
      return 0;
    if ((w = w->RPointer) == NULL)
      return 0;    
  }
}



DoubleLinkedList* SimpleSkipList::ReturnNextLeft()
{ if (CurrentLeft == NULL) return NULL;
  if (CurrentLeft->LPointer != Left)
    return CurrentLeft = CurrentLeft->LPointer;
  return NULL; 
}


DoubleLinkedList* SimpleSkipList::ReturnNextRight()
{ if (CurrentRight == NULL) return NULL;
  if (CurrentRight->RPointer != Right)
    return CurrentRight = CurrentRight->RPointer;
  return NULL; 
}

DoubleLinkedList* SimpleSkipList::ReturnCurrentRight()
{ return CurrentRight; }

DoubleLinkedList* SimpleSkipList::ReturnCurrentLeft()
{ return CurrentLeft; }


DoubleLinkedList* SimpleSkipList::ReturnRight()
{ CurrentLeft = Right->LPointer;
  return Right->LPointer;
}

DoubleLinkedList* SimpleSkipList::ReturnLeft()
{ CurrentRight = Left->RPointer;
  return Left->RPointer;
}

void SimpleSkipList::DeleteElement(DoubleLinkedList* Pointer,
                                   long TIndex)
{
  if (Pointer->LPointer->EntryIndex == Pointer->EntryIndex)
  { ;
  }
  else
  {if (Pointer->RPointer->EntryIndex == Pointer->EntryIndex)
    { if (SkipTable[Pointer->EntryIndex] == Pointer)
        SkipTable[Pointer->EntryIndex] = Pointer->RPointer;
    }
    else
    { if (Pointer->EntryIndex != NULL)
        SkipTable[Pointer->EntryIndex] = NULL;
    } 
  }

  Pointer->LPointer->RPointer = Pointer->RPointer;
  Pointer->RPointer->LPointer = Pointer->LPointer;
  Pointer->LPointer->TriangleIndex = TIndex;
//  delete Pointer;
  InsertIntoPool(Pointer);
  Pointer = NULL;
}


void SimpleSkipList::DeleteInterval(DoubleLinkedList* LeftIn, 
                                    DoubleLinkedList* RightIn,
                                    long TIndex)
{
  if (LeftIn == RightIn)
    DeleteElement(LeftIn, TIndex);
  else
  { DoubleLinkedList* a;
    DoubleLinkedList* l1 = LeftIn->LPointer;
    DoubleLinkedList* r1 = RightIn->RPointer; 

    while ((LeftIn != NULL) && //(LeftIn != RightIn->RPointer))
                               (LeftIn != r1))
    {

      if (LeftIn->LPointer->EntryIndex == LeftIn->EntryIndex)
      { //the pointer into the list will not be changed
        ;
      }
      else
      { if (LeftIn->RPointer->EntryIndex == LeftIn->EntryIndex)
        { if (SkipTable[LeftIn->EntryIndex] == LeftIn)
            SkipTable[LeftIn->EntryIndex] = LeftIn->RPointer;
        }
        else
          SkipTable[LeftIn->EntryIndex] = NULL;
      }
      a = LeftIn;
      LeftIn = LeftIn->RPointer;
     // delete a;
     InsertIntoPool(a);
    }

    l1->RPointer = r1;
    r1->LPointer = l1;
    l1->TriangleIndex = TIndex;
  }
}

void SimpleSkipList::DeleteLeft()
{ DoubleLinkedList* Pointer = Left->RPointer;

  if (SkipTable[Pointer->EntryIndex] == Pointer)
    if (Pointer->LPointer->EntryIndex == Pointer->EntryIndex)
      SkipTable[Pointer->EntryIndex] = Pointer->LPointer;
    else   
      SkipTable[Pointer->EntryIndex] = NULL;

  DoubleLinkedList* a = Left->RPointer;
  DoubleLinkedList* next = a->RPointer;

  Left->RPointer = next;
  next->LPointer = Left;
 // delete a;
  InsertIntoPool(a);
}

void SimpleSkipList::DeleteRight()
{ DoubleLinkedList* Pointer = Right->LPointer;

  if (SkipTable[Pointer->EntryIndex] == Pointer)
      SkipTable[Pointer->EntryIndex] = NULL;

  DoubleLinkedList* a = Right->LPointer;
  DoubleLinkedList* next = a->LPointer;

  Right->LPointer = next;
  next->RPointer = Right;
//  delete a;
  InsertIntoPool(a);
}

void SimpleSkipList::PrintEntries()
{ FILE* f = fopen("SkipEntries.txt", "wt");

  for (int i = 0; i < NoOfCells; i++)
    if (i == 0)
      fprintf(f, "%d -xMax  \n", i);
    else
    { if (SkipTable[i] == NULL)
        fprintf(f, "%d NULL \n", i);
      else
        fprintf(f, "%d %lf %lf \n", i, SkipTable[i]->key, SkipTable[i]->y);
    }

  fclose(f);
}


void SimpleSkipList::PrintAFLow()
{ FILE* f = fopen("AFLow.txt", "wt");
  DoubleLinkedList* w= Left->RPointer;
  

  while (w->RPointer != NULL)
  { fprintf(f, "v:%ld x:%lf y:%lf ti:%ld \n",
                w->VertexNo, w->key, w->y, w->TriangleIndex);
    w = w->RPointer;
  }

  fclose(f);
}





void InitAFHigh()
{ DCLLeft = DCLRight = NULL;}


void FreeAFHigh()
{ while (DCLLeft != NULL)
  { DCLRight = DCLLeft;
    DCLLeft = DCLLeft->RPointer;
    delete DCLRight;
  }
  DCLLeft = DCLRight = NULL;
}


void AFHighAppendLeft(double x, double y, long v, long TIndex)
{
  DoubleLinkedList* a = new DoubleLinkedList;
  a->EntryIndex = NOINDEX;
  a->key = x;
  a->y = y;
  a->VertexNo = v;
  a->TriangleIndex = NOINDEX;
  a->RPointer = a->LPointer = NULL;

  if (DCLLeft == NULL)
    DCLLeft = DCLRight = a;
  else
  { a->LPointer = NULL;
    a->RPointer = DCLLeft;
    DCLLeft->LPointer = a;
    DCLLeft = a;
  }

  DCLLeft->TriangleIndex = TIndex;
  DCLLeft->Next = NULL;
}

void AFHighAppendRight(double x, double y, long v, long TIndex)
{
  DoubleLinkedList* a = new DoubleLinkedList;
  a->EntryIndex = NOINDEX;
  a->key = x;
  a->y = y;
  a->VertexNo = v;
  if (TIndex == 0) // first triangle
    a->TriangleIndex = NOINDEX;
  a->RPointer = a->LPointer = NULL;

  if (DCLRight == NULL)
    DCLLeft = DCLRight = a;
  else
  { a->LPointer = DCLRight;
    a->RPointer = NULL;
    DCLRight->RPointer = a;
    DCLRight = a;
  }
  if (DCLRight->LPointer != NULL)
    DCLRight->LPointer->TriangleIndex = TIndex;
  else 
    DCLRight->TriangleIndex = NOINDEX;
}

void DeleteFromLeft(DoubleLinkedList* Pointer)
{ if (Pointer != NULL)
  { DoubleLinkedList* a;
    DoubleLinkedList* w = DCLLeft; 
    
    while ((w != Pointer) && (w != NULL))
    { a = w;
      w = w->RPointer;
      delete a;
    }
    DCLLeft = w->RPointer;
    delete w;
    DCLLeft->LPointer = NULL;
  }
}

void DeleteFromRight(DoubleLinkedList* Pointer)
{ if (Pointer != NULL)
  { DoubleLinkedList* a;
    DoubleLinkedList* w = DCLRight; 
    
    while ((w != Pointer) && (w != NULL))
    { a = w;
      w = w->LPointer;
      delete a;
    }
    DCLRight = w->LPointer;
    delete w;
    DCLRight->RPointer = NULL;
  }
}


void UpdateTIndexLow(long v1, long v2, long TIndexNew)
{ DoubleLinkedList* w = DCLLeft; 
  DoubleLinkedList* w1 = w->RPointer;
  int EndFlag = 0;
  while ((w1 != NULL) && !EndFlag)
  { if ( ((w1->VertexNo == v1) && (w->VertexNo == v2)) ||
         ((w1->VertexNo == v2) && (w->VertexNo == v1)) )
    { EndFlag = 1;
      w->TriangleIndex = TIndexNew;
    }
    else
    { w = w1;
      w1 = w1->RPointer;
    }
  } 
}

DoubleLinkedList* ReturnDCLLeft() {return DCLLeft;}
DoubleLinkedList* ReturnDCLRight() {return DCLRight;}

void PrintDCLList()
{
  FILE* f = fopen("DCL.txt", "wt");

  DoubleLinkedList* w = DCLLeft;

  while (w != NULL)
  { fprintf(f, "%lf %lf %ld \n", w->key, w->y, w->TriangleIndex);
    w = w->RPointer;
  }
  fclose(f);
}