// SimpleSkipList.h: interface for the SimpleSkipList class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SIMPLESKIPLIST_H__ED77BC84_FCDF_4A19_B2C6_EF7B2851008D__INCLUDED_)
#define AFX_SIMPLESKIPLIST_H__ED77BC84_FCDF_4A19_B2C6_EF7B2851008D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define NOINDEX -1

#define LEFT -1
#define RIGHT 1
#define MIDDLE 0
#define ON_EDGE 2
#define VERTICALLY_COLINEAR -3
#define COLINEAR -4
#define POINTS_THE_SAME -2
#define GAP_BORDER 1.0


typedef struct DoubleLinkedList1 
{ double key, y;
  long TriangleIndex; // pointer to triangle
  long EntryIndex;
  long VertexNo;      // vertexIndex

  DoubleLinkedList1* RPointer;
  DoubleLinkedList1* LPointer;

  DoubleLinkedList1* Next; // for making the stack
} DoubleLinkedList;


class SimpleSkipList  
{
private:
  long NoOfCells;
  double XMax, XMin, Size;

  DoubleLinkedList* Left;
  DoubleLinkedList* Right;
  DoubleLinkedList* CurrentLeft;
  DoubleLinkedList* CurrentRight;
  DoubleLinkedList** SkipTable;

  void AddAfter(DoubleLinkedList* OldLeft, DoubleLinkedList* NewE);
  void AddBefore(DoubleLinkedList* walker, DoubleLinkedList* NewEl);

  DoubleLinkedList* ReturnFromPool();
  void InsertIntoPool(DoubleLinkedList* w);

  DoubleLinkedList1* Pool;
public:
  SimpleSkipList(long np, long fraction, double min, double max);
  virtual ~SimpleSkipList();

  int Insert(double x, double y, long v, long TIndex);
 
  int AppendAtTheLeft(double x, double y, long v, long TIndex);
  int AppendAtTheRight(double x, double y, long v, long TIndex);
  int ReturnInterval(double x, double y, DoubleLinkedList** Left, DoubleLinkedList** Right);
  void DeleteInterval(DoubleLinkedList* LeftIn, 
                      DoubleLinkedList* RightIn,
                      long TIndex);
  void DeleteElement(DoubleLinkedList* Pointer,
                     long TIndex);
  void DeleteLeft();
  void DeleteRight();
  

  DoubleLinkedList* ReturnNextRight(); 
  DoubleLinkedList* ReturnNextLeft(); 

  DoubleLinkedList* ReturnCurrentRight();
  DoubleLinkedList* ReturnCurrentLeft();

  DoubleLinkedList* ReturnRight(); 
  DoubleLinkedList* ReturnLeft(); 

  //void UpdateTIndex(double x, double y, long ti);
  int UpdateTIndex(double x, long i, long j, long ti);

  void ChangeTriangleIndex(long vertex,long TIndex);

  int ReturnSkipTable0Status() 
  { if (SkipTable[0] == NULL) return 1;
    return 0;
  }

  void PrintEntries();
  void PrintAFLow();

};


// functions for controling double connected list
void InitAFHigh();
void FreeAFHigh();
void AFHighAppendLeft(double x, double y, long v, long TIndex);
void AFHighAppendRight(double x, double y, long v, long TIndex);
void DeleteFromLeft(DoubleLinkedList* Pointer);
void DeleteFromRight(DoubleLinkedList* Pointer);
DoubleLinkedList* ReturnDCLLeft();
DoubleLinkedList* ReturnDCLRight();

void PrintDCLList();

#endif // !defined(AFX_SIMPLESKIPLIST_H__ED77BC84_FCDF_4A19_B2C6_EF7B2851008D__INCLUDED_)
