// SkipList.cpp: implementation of the SkipList class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyTriangF.h"
#include "SkipList.h"
#include <stdio.h>


#include "globals.h"
#include "utility.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

static int SkipValue = 10;
static int HalfSkipValue = 5;
static SkipList** MyStack;
static int StackPointer = 0;


SkipList* InitSkipList(int sv)
{ SkipValue = sv;

// level 1
  HalfSkipValue = 1 + (int)(sv * 0.5);


// level 2

  SkipList* as = new SkipList;
  as->DeleteFlag = 0;
  as->down = NULL;
  as->key = -MAX_DOUBLE;
  as->terminalPointer = NULL;
  as->Counter = 0;

  SkipList* bs = new SkipList;
  bs->DeleteFlag = 0; 
  bs->down = bs->forward = NULL;
  bs->key = MAX_DOUBLE;
  bs->Counter = 0;
  bs->terminalPointer = NULL;
  as->forward = bs;


  SkipList* s = new SkipList;
  as->down = s;
  s->DeleteFlag = 0;
  s->down = NULL;
  s->key = -MAX_DOUBLE;
  s->Counter = 0;

  SkipList* s1 = new SkipList;
  s1->DeleteFlag = 0; 
  s1->down = s1->forward = NULL;
  s1->key = MAX_DOUBLE;
  s1->Counter = 0;
  s1->terminalPointer = NULL;
  s->forward = s1;




  TerminalNode* a = new TerminalNode;
  a->DeleteFlag = 0;
  a->left = NULL;
  a->x = -MAX_DOUBLE;   a->y = -MAX_DOUBLE;
  a->TriangleIndex = -1;
  s->terminalPointer = a;


  TerminalNode* b = new TerminalNode;
  b->DeleteFlag = 0;
  b->left = a;
  a->right = b;
  b->right = NULL;
  b->x = MAX_DOUBLE;   b->y = MAX_DOUBLE;
  b->TriangleIndex = -1;

  MyStack = new SkipList*[1000];
  StackPointer = 0;

  return as;
}


void FreeSkipList(SkipList* walkerSl)
{ int flag = 1;
  StackPointer = 0;
  while (flag)
  { MyStack[StackPointer++] = walkerSl;

    if (walkerSl->terminalPointer != NULL)
      flag = 0;
    else
      walkerSl = walkerSl->down;
  }

  TerminalNode* t1 = walkerSl->terminalPointer;
  TerminalNode* t2;

  while (t1 != NULL)
  { t2 = t1;
    t1 = t1->right;
    delete t2;
  }

  SkipList* s1;
  SkipList* s2;
  while (StackPointer != 0)
  { s1 = MyStack[--StackPointer];
    while (s1 != NULL)
    { s2 = s1;
      s1 = s1->forward;
      delete s2;
    }
  }
  delete[] MyStack;
}
 


int InsertIntoSL(SkipList** skl, double x, double y, long ti)
{ SkipList* walkerSl = (*skl)->down;
 
  TerminalNode* tn;
  TerminalNode* w;
  TerminalNode* w1;

  StackPointer = 0;

  // move towards terminals
  int flag = 1;
  while (flag)
  { while (walkerSl->forward->key < x) 
      walkerSl = walkerSl->forward;

    MyStack[StackPointer++] = walkerSl;

    if (walkerSl->terminalPointer != NULL)
      flag = 0;
    else
      walkerSl = walkerSl->down;
  }
  walkerSl->Counter++;
  w = walkerSl->terminalPointer;
  while (w->x < x) 
    w = w->right;

  if (TheSamePoints2D(x, y, w->x, w->y))
    return 0;
 
  tn = new TerminalNode;
  tn->DeleteFlag = 1;
  tn->TriangleIndex = ti;
  tn->x = x;
  tn->y = y;
  
  w1 = w->left;
  w1->right = tn;
  tn->left = w1;
  w->left = tn;
  tn->right = w;

  if (walkerSl->Counter >= SkipValue) 
  // skip list must be reconfigured.
  { w = walkerSl->terminalPointer;
    walkerSl->Counter = HalfSkipValue;
    for (int i = 0; i < walkerSl->Counter; i++) 
    { w = w->right; 
    }
    walkerSl->Counter--;

    SkipList* newSL;
    SkipList* sl = MyStack[--StackPointer];    

    newSL = new SkipList;
    newSL->DeleteFlag = 0;
    newSL->Counter = SkipValue-HalfSkipValue;
    newSL->terminalPointer = w;
    newSL->key = w->x;
    newSL->forward = sl->forward;
    sl->forward = newSL;

    flag = 1;
    SkipList* slOld;
    while (flag) // let us check higher levels
    { slOld = sl; 
      StackPointer--;

      if (StackPointer > 0)
      { sl = MyStack[StackPointer];
        sl->Counter++;
        if (sl->Counter > SkipValue)
        { for (int i = 0; i < HalfSkipValue; i++)
            slOld = slOld->forward;

          newSL = new SkipList;
          newSL->DeleteFlag = 0;
          newSL->Counter = SkipValue-HalfSkipValue+1;
          newSL->down = slOld;
          newSL->key = slOld->key;
          newSL->forward = sl->forward;
          sl->forward = newSL;
        }
        else flag = 0;
      }
      else
        flag = 0;
    }
    // at the end the new upper level is added if neccesary
    walkerSl = *skl;
    int Counter = -2;
    while (walkerSl != NULL)
    { Counter++;
      walkerSl = walkerSl->forward;
    }
    if (Counter >= SkipValue)
    { //newSL = GenerateTop();
      //newSL->down = *skl;
      //*skl = newSL;
    }
  }
  return 1;
}




int DeleteFromSL(SkipList* skl, TerminalNode* t1,
                                TerminalNode* t2)
// function erase terminal nodes from the Terminal node 
// t1 to t2 from skip list; t1 and t2 are deleted
{ TerminalNode* t;

  if ((t1 == NULL) || (t2 == NULL))
    return 0;
  if (t1->x > t2->x)
  { t = t1; t1 = t2; t2 = t; }

  // mark the path to terminal node at first;
  // move towards terminals
 
  SkipList* walkerSl = skl;
  int flag = 1;
  StackPointer = 0;
  while (flag)
  { while (walkerSl->forward->key < t1->x) walkerSl = walkerSl->forward;

    MyStack[StackPointer++] = walkerSl;

    if (walkerSl->terminalPointer != NULL)
      flag = 0;
    else
      walkerSl = walkerSl->down;
  }
  walkerSl->Counter++;
  TerminalNode* w = walkerSl->terminalPointer;
  while (w != t1) w = w->right;

  // Mark the terminal which will be deleted
  while (w != t2)
  { w->DeleteFlag = 1;
    w = w->right;
  }
  // mark all nonterminals pointing on deleted terminal for
  // deletion

  int StoreStackPointer = StackPointer;
  SkipList* wsl = MyStack[--StackPointer];
  while ((wsl != NULL) && (wsl->terminalPointer->DeleteFlag == 0))
  { wsl->DeleteFlag = 1;
    wsl = wsl->forward;
  }

  while (StackPointer != 0)
  { wsl = MyStack[--StackPointer];
    while ((wsl != NULL) && (wsl->down->DeleteFlag == 1))
    { wsl->DeleteFlag = 1;
      wsl = wsl->forward;
    }
  }

  // now we delete all nodes marked for deleting
 
  TerminalNode* StoreLeft = t1->left;
  t1->left->right = t2->right;
  t2->right->left = t1->left;
  t2->right = NULL;

  while (t1 != NULL)  // clean the terminal nodes
  { t2 = t1;
    t1 = t1->right;
    delete t2;
  }
  t1 = t2 = NULL;

  // clean the inner nodes
  SkipList* w1;
  while (StackPointer != 0)
  { w1 = MyStack[--StackPointer];
    wsl = w1->forward;

    while ((wsl != NULL) && (wsl->DeleteFlag == 1))
    { w1->forward = wsl->forward;
      delete wsl;
      wsl = w1->forward;
    }
  }


 // let do the bance at the end;
 
  StackPointer = StoreStackPointer;
  w1 =MyStack[--StackPointer];
  t1 = w1->terminalPointer;
  int count = 0;
  while ((t1 != NULL) && (t1 != w1->terminalPointer))
  { count++;
    t1 = t1->right;
  }
  if (count > SkipValue)
  { int c1 = (int)(0.5 * count);

    t1 = w1->terminalPointer;
    for (int i = 0; i < c1; i++) t1 = t1->right;

    SkipList* newSL = new SkipList;
    newSL->DeleteFlag = 0;
    newSL->Counter = count - c1;
    newSL->terminalPointer = t1;
    newSL->key = t1->x;
    newSL->forward = w1->forward;
    w1->forward = newSL;
    w1->Counter = c1;
  }

  SkipList* w2;
  while (StackPointer != NULL)
  { w1 = MyStack[StackPointer];
        
    w2 = w1->down;
    count = 0;
    while (w2 != w1->forward->down)
    { count++;
      w2 = w2->forward;
    }

    if (count >= SkipValue)
    { int c1 = (int)(0.5 * count);

      SkipList* w3 = w1;
      for (int i = 0; i < c1; i++) w3 = w3->forward;

      SkipList* newSL = new SkipList;
      newSL->DeleteFlag = 0;
      newSL->Counter = count - c1;
      newSL->down = w3;
      newSL->terminalPointer = NULL;
      newSL->key = w3->key;
      newSL->forward = w1->forward;
      w1->forward = newSL;
      w1->Counter = c1;
    }
  }
  return 1;
}


int FindInterval(SkipList* skl, double x, 
            TerminalNode** TLeft,
            TerminalNode** TRight)
{ SkipList* walkerSl = skl;
 
  // move towards terminals
  int flag = 1;
  while (flag)
  { while (walkerSl->forward->key < x) 
      walkerSl = walkerSl->forward;

    if (walkerSl->terminalPointer != NULL)
      flag = 0;
    else
      walkerSl = walkerSl->down;
  }

  *TLeft = walkerSl->terminalPointer;
  while ((*TLeft)->right->x < x)
   *TLeft = (*TLeft)->right;
  
  *TRight = (*TLeft)->right;

  return 1;
}

void PrintOutSkipList(SkipList* skl, char* FName)
{ FILE* f = fopen(FName, "wt");

  SkipList* w = skl;
  SkipList* s = skl;
   int level = 1;
  int endflag = 1;
  do 
  { fprintf(f,"Skip List Level %d \n", level++);
    
    w = s;
     while (w != NULL) 
     { if ((w->key != MAX_DOUBLE) && (w->key != -MAX_DOUBLE))
         fprintf(f, "%lf(%d) ,", w->key, w->Counter);
       else fprintf(f, "(%d) ,", w->Counter);
       w = w->forward;
     }
     fprintf(f, " \n");
     if (s->down != NULL) 
       s = s->down;
     else 
       endflag = 0;
  } while (endflag == 1);

  TerminalNode* t = s->terminalPointer;
  while (t != NULL)
  { if ((t->left != NULL) && (t->right != NULL))
      fprintf(f, "(%lf,%lf)", t->x, t->y);
    t = t->right;
  } 


  fclose(f);
}


