// SweepTriangle.cpp: implementation of the SweepTriangle class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyTriangF.h"
#include "SweepTriangle.h"
//#include "SkipList.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// global variables

static VertexStruct* Vertices = NULL;
long NoOfVertices = 0;

static SimpleSkipList* AFLow = NULL; // this AF follows sweep-line
static SimpleSkipList* AFHigh = NULL; // this is the same as convex hull

double C[3][3];

BaseTriangle* TriangArray;
long NextTriangPosition = 0;


//#define PRINTOUT


//SkipList* SLPointer = NULL;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
int InitSweepDT(long n, double* xarr, double* yarr)
{  

  if (n < 2) return 0;

  double min = MAX_DOUBLE;
  double max = -MAX_DOUBLE;
  Vertices = new VertexStruct[n];
  for (long i = 0; i < n; i++)
  { Vertices[i].x = xarr[i];
    Vertices[i].y = yarr[i];
    Vertices[i].Index = i;
    if (xarr[i] < min) min = xarr[i];
    if (xarr[i] > max) max = xarr[i];
  }

  int iLo = 0;
  int iHi = n-1;
  QSort(iLo, iHi);
 
  if (Equal(Vertices[0].y, Vertices[1].y))
  { if (Vertices[0].x < Vertices[1].x)
    { VertexStruct a = Vertices[0];
      Vertices[0] = Vertices[1];
      Vertices[1] = a;
    }
  }




  TriangArray = new BaseTriangle[2*n]; 

  NoOfVertices = n;

  long BottomFlag = 0;
  if (Equal(Vertices[0].y, Vertices[2].y))
  { if ((BottomFlag = BottomTriangles()) == 0)
      return NoOfVertices; // all vertices are collinear

    AFLow = new SimpleSkipList(NoOfVertices, 2, min, max);
    InitAFHigh();
    InitAFBottom(BottomFlag);



#ifdef PRINTOUT

FILE* f = fopen("VERTICES.TXT","wt");
for (i = 0; i < n; i++)
  fprintf(f, "%ld %lf %lf %ld \n", i, Vertices[i].x, 
                                      Vertices[i].y, 
                                      Vertices[i].Index);
fclose(f);





int n1, n2;
n1 = 150;
n2 = 330;

if (n1 < n)
{

int kk = 0;

for (i = n1; i<n2; i++)
  if ((Vertices[i].x < -30) && (Vertices[i].x > -90 )) kk++;



FILE* g = fopen("V.PNT","wt");

fprintf(g, "%d\n",kk);
for (i = n1; i < n2; i++)
  if ((Vertices[i].x < -30) && (Vertices[i].x > -90 ))
    fprintf(g, "%lf, %lf\n", Vertices[i].x, Vertices[i].y);
fclose(g);
}




#endif


    return BottomFlag+1;
  }
  else
  { FirstTriangle();
    AFLow = new SimpleSkipList(NoOfVertices, 2, min, max);
    InitAFHigh();
    InitAF(); 


#ifdef PRINTOUT

FILE* f = fopen("VERTICES.TXT","wt");
for (i = 0; i < n; i++)
  fprintf(f, "%ld %lf %lf %ld \n", i, Vertices[i].x, 
                                      Vertices[i].y, 
                                      Vertices[i].Index);
fclose(f);


int n1, n2;
n1 = 200;
n2 = 250;

if (n1 < n)
{

FILE* g = fopen("V.PNT","wt");
fprintf(g, "%d\n",n2-n1);
for (i = 200; i < 250; i++)
  fprintf(g, "%lf, %lf\n", Vertices[i].x, 
                           Vertices[i].y);
fclose(g);
}


#endif


    return 3;
  }
}

void FinalSweepDT()
{
  if (Vertices != NULL)
  { delete[] Vertices;
    Vertices = NULL;
  }


  if (TriangArray != NULL)
  { delete[] TriangArray;
    TriangArray = NULL;
  }

  delete AFLow;
  AFLow = NULL;
  FreeAFHigh();
  AFLow = AFHigh = NULL;
}


void QSort(long iLo, long iHi) 
// quick sort of the array Vertices regarding y coordinate.
{
  long Lo, Hi;
  double Mid;
  VertexStruct T;

  Lo = iLo;
  Hi = iHi;
  Mid = Vertices[(Lo+Hi)/2].y;
  do
  {
    while (Vertices[Lo].y < Mid)
        Lo++;
    while (Vertices[Hi].y > Mid)
        Hi--;
    if (Lo <= Hi)
    {
      T = Vertices[Lo];
      Vertices[Lo] = Vertices[Hi];
      Vertices[Hi] = T;
      Lo++;
      Hi--;
    }
  }
  while (Lo <= Hi);

  if (Hi > iLo)
    QSort(iLo, Hi);
  if (Lo < iHi)
    QSort(Lo, iHi);
}

int InitAF()
{ long i = TriangArray[0].i;
  long j = TriangArray[0].j;
  long k = TriangArray[0].k;

  if ( ((Vertices[j].x > Vertices[k].x) &&
        (Vertices[j].x < Vertices[i].x)) ||
       ((Vertices[j].x < Vertices[k].x) &&
        (Vertices[j].x > Vertices[i].x)))
  { 
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
        
    AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
    return 1;   
  }

    

  

  if (Vertices[i].x > Vertices[j].x) 
  {   
    if (Vertices[k].x < Vertices[i].x)
    { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
      AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
      AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
      return 1;
    }
    else
    { AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);
      AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
      AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, 0);
      return 1;
    }
  }

  if (Vertices[k].x < Vertices[i].x)
  { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

    AFHighAppendRight(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    return 1;
  }
  else
  { AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendLeft(Vertices[i].x, Vertices[i].y, i, 0);
    AFHighAppendRight(Vertices[j].x, Vertices[j].y, j, 0);
    return 1;
  }
}

void InitAFBottom(long n )
{
  if ((Vertices[0].x <= Vertices[n].x) && 
      (Vertices[n].x <= Vertices[n-1].x) )
  { AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
    AFLow->AppendAtTheRight(Vertices[n].x, Vertices[n].y, n, n-2);
    AFLow->AppendAtTheRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
    AFLow->UpdateTIndex(Vertices[n].x, n, n-1, n-2);
    AFLow->UpdateTIndex(Vertices[0].x, 0, n, 0);

    AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
    AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
  }
  else // AF contains only two vertices
  { if (Vertices[n].x < Vertices[0].x)
    { AFLow->Insert(Vertices[n].x, Vertices[n].y, n, n-2);
      AFLow->AppendAtTheRight(Vertices[n-1].x, Vertices[n-1].y, n-1, n-2);
      AFLow->UpdateTIndex(Vertices[n-1].x, n, n-1, n-2); 
      
      AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
      AFHighAppendLeft(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
      AFHighAppendLeft(Vertices[n].x, Vertices[n].y, n, NOINDEX);
    } 
    else
    { AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
      AFLow->AppendAtTheRight(Vertices[n].x, Vertices[n].y, n, 0);
      AFLow->UpdateTIndex(Vertices[n].x, 0, n, 0); 

      AFHighAppendLeft(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
      AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
      AFHighAppendRight(Vertices[n].x, Vertices[n].y, n, NOINDEX);


    }
  }
}

void FirstTriangle()
{ NextTriangPosition = 0; 
  TriangArray[NextTriangPosition].i = 2;
  TriangArray[NextTriangPosition].j = 0;
  TriangArray[NextTriangPosition].k = 1;

/*  if (Vertices[1].x > Vertices[2].x)
  {
    TriangArray[NextTriangPosition].k = 2;
    TriangArray[NextTriangPosition].i = 1;
  }
*/

  if (SignedArea(Vertices[0].x, Vertices[0].y, 
                 Vertices[2].x, Vertices[2].y,
                 Vertices[1].x, Vertices[1].y) < 0)
  {
    TriangArray[NextTriangPosition].i = 1;
    TriangArray[NextTriangPosition].k = 2;
  }

  TriangArray[NextTriangPosition].ij = 
  TriangArray[NextTriangPosition].jk =
  TriangArray[NextTriangPosition].ki = NONE;

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;

  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

  NextTriangPosition++;

}



int BottomTriangles()
// more vertices is on the first scan line
{ long i = 1;
  long n = 0;

  while ((i < NoOfVertices) && Equal(Vertices[i].y, Vertices[i-1].y))
  { i++;
    n++;
  }
  if (n == NoOfVertices)
    return 0;

  long j;
  VertexStruct VS;
  for (i = 0; i < n; i++)
    for (j = i+1; j < n+1; j++)
    { if (Vertices[j].x < Vertices[i].x)
      { VS = Vertices[j];
        Vertices[j] = Vertices[i];
        Vertices[i] = VS;
      }
    }

  n++;
  NextTriangPosition = 0;
  for (i = 1; i < n; i++)
  { TriangArray[NextTriangPosition].i = n;
    TriangArray[NextTriangPosition].j = i;
    TriangArray[NextTriangPosition].k = i-1;

    TriangArray[NextTriangPosition].ij = NOINDEX;
    TriangArray[NextTriangPosition].jk = NOINDEX;
    TriangArray[NextTriangPosition].ij = i;
    if (i == 1)
      TriangArray[NextTriangPosition].ki = NONE;
    else
      TriangArray[NextTriangPosition].ki = NextTriangPosition-1;
    

    TriangArray[NextTriangPosition].Eij = 
    TriangArray[NextTriangPosition].Ejk =
    TriangArray[NextTriangPosition].Eki = FALSE;

    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

    NextTriangPosition++;
  }
  TriangArray[NextTriangPosition-1].ij = NOINDEX;
  return n;
}


void SweepDT(long StartVertex)
// this is the main triangulation function performing DT
// using sweep-line approach
{ long i;
  DoubleLinkedList* LeftAF;
  DoubleLinkedList* RightAF;

//NoOfVertices = 320;
//NoOfVertices = 50;
#ifdef PRINTOUT
AFLow->PrintEntries();
AFLow->PrintAFLow();
PrintTriangulation();
#endif

  for (i = StartVertex; i < NoOfVertices; i++)
  { 
if (i == 41)
{
int  uuu = 8;
}


      switch (AFLow->ReturnInterval(Vertices[i].x,
                                    Vertices[i].y, 
                                    &LeftAF, 
                                    &RightAF))
      { case MIDDLE: 
          SolveMiddleCase(i, LeftAF, RightAF); 
          break;
        case LEFT: 
          SolveLeftCase(i); 
          break;
        case RIGHT: 
          SolveRightCase(i);
          break;
        case ON_EDGE:
          SolveOnEDGE(i, LeftAF, RightAF); // se napaka pri sosedih
         break;           
      }
#ifdef PRINTOUT
AFLow->PrintEntries();
AFLow->PrintAFLow();
PrintTriangulation();
PrintDCLList();
#endif
 
  }
}


void SetNeighbour(long T1, long v1, long v2, long TIndex)
{ if ( ((TriangArray[T1].i == v1) && (TriangArray[T1].j == v2)) ||
       ((TriangArray[T1].i == v2) && (TriangArray[T1].j == v1)))
    TriangArray[T1].ij = TIndex;
  else
  { if ( ((TriangArray[T1].j == v1) && (TriangArray[T1].k == v2)) ||
         ((TriangArray[T1].j == v2) && (TriangArray[T1].k == v1)))
      TriangArray[T1].jk = TIndex;
    else
     TriangArray[T1].ki = TIndex;
  }
}


void InsertNewTriangle(long NewVertexIndex, DoubleLinkedList* lt, DoubleLinkedList* rt)
{ long OldT = lt->TriangleIndex;
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = rt->VertexNo;
  long k = TriangArray[NextTriangPosition].k = lt->VertexNo;

  TriangArray[NextTriangPosition].ij = 
  TriangArray[NextTriangPosition].ki = NONE;
  TriangArray[NextTriangPosition].jk = OldT;

  SetNeighbour(OldT, k, j,NextTriangPosition);
  CheckBorderFlag(OldT);
/*
  if ((TriangArray[OldT].i == k) && 
      (TriangArray[OldT].j == j))
    TriangArray[OldT].ij = NextTriangPosition;
  else
    if ((TriangArray[OldT].i == j) && 
        (TriangArray[OldT].k == k))
      TriangArray[OldT].ki = NextTriangPosition;
*/
  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;

  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}

void CloseNewTriangleLeft(long NewVertexIndex, 
                          DoubleLinkedList* LeftP, 
                          DoubleLinkedList* NextLeftP)
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = LeftP->TriangleIndex;
  TriangArray[NextTriangPosition].ki = NOINDEX;
  TriangArray[NextTriangPosition].jk = NextLeftP->TriangleIndex;

  SetNeighbour(LeftP->TriangleIndex, j, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(LeftP->TriangleIndex);

  SetNeighbour(NextLeftP->TriangleIndex, j, k, NextTriangPosition);
  CheckBorderFlag(NextLeftP->TriangleIndex);

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}


void CloseNewTriangleLeftMiddle(long NewVertexIndex, 
                          DoubleLinkedList* LeftP, 
                          DoubleLinkedList* NextLeftP)
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = LeftP->TriangleIndex;
  TriangArray[NextTriangPosition].jk = NextLeftP->TriangleIndex;
  TriangArray[NextTriangPosition].ki = NOINDEX;

  SetNeighbour(NextLeftP->TriangleIndex, k, j, NextTriangPosition);
  CheckBorderFlag(NextLeftP->TriangleIndex);

  SetNeighbour(LeftP->TriangleIndex, j, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(LeftP->TriangleIndex);

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}


void CheckBorderFlag(int TIndex)
{ 
  if ((TriangArray[TIndex].ij != NOINDEX) &&
      (TriangArray[TIndex].jk != NOINDEX) &&
      (TriangArray[TIndex].ki != NOINDEX))
    TriangArray[TIndex].BorderTriangleFlag = FALSE;
  else
    TriangArray[TIndex].BorderTriangleFlag = TRUE;
}


void CloseNewTriangleLeftDown(long NewVertexIndex, 
                              long TIndex,
                              DoubleLinkedList* LeftP, 
                              DoubleLinkedList* NextLeftP)
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX;
  TriangArray[NextTriangPosition].ki = TIndex;
  TriangArray[NextTriangPosition].jk = LeftP->TriangleIndex;
//  CheckBorderFlag(NextTriangPosition);
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

  if (TIndex != NOINDEX)
  { SetNeighbour(TIndex, TriangArray[NextTriangPosition].k, 
                         TriangArray[NextTriangPosition].i, 
                         NextTriangPosition); //!!!!
    CheckBorderFlag(TIndex);
  }

  SetNeighbour(LeftP->TriangleIndex, k, j,NextTriangPosition);
  CheckDT(LeftP->TriangleIndex, NextTriangPosition);


  CheckBorderFlag(LeftP->TriangleIndex);

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
}



void CloseNewTriangleRightDown(long NewVertexIndex, 
                              long TIndex,
                              DoubleLinkedList* RightP, 
                              DoubleLinkedList* NextRightP)
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = RightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextRightP->VertexNo;

  TriangArray[NextTriangPosition].ij = TIndex;
  TriangArray[NextTriangPosition].jk = RightP->TriangleIndex ;
  TriangArray[NextTriangPosition].ki = NOINDEX;
  CheckBorderFlag(NextTriangPosition);

  if (TIndex != NOINDEX) 
  { 
    SetNeighbour(TIndex, TriangArray[NextTriangPosition].j, 
                         TriangArray[NextTriangPosition].i, 
                         NextTriangPosition); //!!!!
    CheckBorderFlag(TIndex);
  }

  SetNeighbour(RightP->TriangleIndex, k, j,NextTriangPosition);
  CheckDT(RightP->TriangleIndex, NextTriangPosition);


  CheckBorderFlag(RightP->TriangleIndex);

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
}


void CloseNewTriangleRight(long NewVertexIndex, 
                           DoubleLinkedList* RightP, 
                           DoubleLinkedList* NextRightP)
{
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = NextRightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = RightP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX; 
  TriangArray[NextTriangPosition].ki = NextTriangPosition - 1;
  long TIndex2 = RightP->TriangleIndex;
  TriangArray[NextTriangPosition].jk = TIndex2;
  CheckBorderFlag(NextTriangPosition);

//  TriangArray[NextTriangPosition - 1].ij = NextTriangPosition;
  SetNeighbour(NextTriangPosition - 1, k, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(NextTriangPosition - 1);

  SetNeighbour(TIndex2, k, j,NextTriangPosition);

/*  if (TriangArray[TIndex2].i == RightP->VertexNo)
    TriangArray[TIndex2].ij = NextTriangPosition;
  else
    TriangArray[TIndex2].ki = NextTriangPosition;
*/
  CheckBorderFlag(TIndex2);
 
  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
}


void CloseNewTriangleRightMiddle(long NewVertexIndex, 
                           DoubleLinkedList* RightP, 
                           DoubleLinkedList* NextRightP)
{
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = NextRightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = RightP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX; 
  TriangArray[NextTriangPosition].jk = RightP->TriangleIndex;
  long Ti = RightP->LPointer->TriangleIndex;
  TriangArray[NextTriangPosition].ki = Ti;

  SetNeighbour(Ti, k, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(Ti);

  SetNeighbour(RightP->TriangleIndex, k, j, NextTriangPosition);
  CheckBorderFlag(RightP->TriangleIndex);
 

  TriangArray[NextTriangPosition].Eij = 
  TriangArray[NextTriangPosition].Ejk =
  TriangArray[NextTriangPosition].Eki = FALSE;
}
 
void   MakeTrianglesToTheRight(int MiddleFlag, double x, double y, long i, 
                               DoubleLinkedList* RightTower)
{
  DoubleLinkedList* NextRightVertex;
  DoubleLinkedList* AFWalker = RightTower;
  int DeleteFlag = 0;
  int TInd;
  if (NextRightVertex = AFLow->ReturnNextRight())
  { 

/*    while ((NextRightVertex !=  NULL) && (SignedArea(x, y, 
                            AFWalker->key, // x
                            AFWalker->y,
                            NextRightVertex->key, 
                            NextRightVertex->y)) > 0) */
    while ((NextRightVertex !=  NULL) && (PositiveSignedArea(x, y, 
                            AFWalker->key, // x
                            AFWalker->y,
                            NextRightVertex->key, 
                            NextRightVertex->y) > 0) )
    { // new triangles have to be added


      if (MiddleFlag)
      { CloseNewTriangleRightMiddle(i, AFWalker, NextRightVertex);
        TInd = AFWalker->TriangleIndex;
        AFLow->DeleteElement(AFWalker, NextTriangPosition);
        DeleteFlag = 0;
      }
      else
      { 
        TInd = AFWalker->TriangleIndex;
        CloseNewTriangleRight(i, AFWalker, NextRightVertex);
        AFLow->DeleteElement(AFWalker, NextTriangPosition);     
        DeleteFlag = 0;

//        DeleteFlag = 1;
      }
// !!!!
      CheckDT(NextTriangPosition, TInd);

      NextTriangPosition++;
      AFWalker = NextRightVertex;
      NextRightVertex = AFLow->ReturnNextRight();

    }

    if (DeleteFlag)
    { if (AFWalker->LPointer != RightTower)
       AFLow->DeleteInterval(RightTower, 
                             AFWalker->LPointer,
                             NextTriangPosition-1);
      else
        AFLow->DeleteElement(RightTower, NextTriangPosition-1);
    }
  }
}


void   MakeTrianglesToTheLeft(int MiddleFlag, double x, double y, long i, 
                               DoubleLinkedList* LeftTower)
{ DoubleLinkedList* NextLeftVertex;
  DoubleLinkedList* AFWalker = LeftTower;
  int DeleteFlag = 0;
  int TInd;
  if (NextLeftVertex = AFLow->ReturnNextLeft())
  { 
    while ((NextLeftVertex !=  NULL) && (SignedArea(x, y, 
                            AFWalker->key, // x
                            AFWalker->y,
                            NextLeftVertex->key, 
                            NextLeftVertex->y)) < 0  ) 

/*    while ((NextLeftVertex !=  NULL) && (NegativeSignedArea(x, y, 
                            AFWalker->key, // x
                            AFWalker->y,
                            NextLeftVertex->key, 
                            NextLeftVertex->y)) < 0  ) */
    { // new triangles have to be added

      if (MiddleFlag)
      { CloseNewTriangleLeftMiddle(i, AFWalker, NextLeftVertex);
        TInd = NextLeftVertex->TriangleIndex;
        AFLow->DeleteElement(AFWalker, NextTriangPosition);
        DeleteFlag = 0;
      }
      else
      { TInd = NextLeftVertex->TriangleIndex;
        CloseNewTriangleLeft(i, AFWalker, NextLeftVertex);   
        AFLow->DeleteElement(AFWalker, NextTriangPosition);     
        DeleteFlag = 0;
      }
// !!!!
      CheckDT(NextTriangPosition, TInd);

      NextTriangPosition++;
      AFWalker = NextLeftVertex;
      NextLeftVertex = AFLow->ReturnNextLeft();
    }

    if (DeleteFlag)
    { if (AFWalker->RPointer != LeftTower)
         AFLow->DeleteInterval(AFWalker->RPointer,LeftTower,                                
                               NextTriangPosition-1);
      else
        AFLow->DeleteElement(LeftTower, NextTriangPosition-1);
    }
  }  
}


int SolveOnEDGE(long nv, DoubleLinkedList* Left, 
                        DoubleLinkedList* Right)
{ long t1 = Left->TriangleIndex;
  long v1 = Left->VertexNo;
  long v2 = Right->VertexNo;

  int Orientation;
  int flag = 0;
  while ((( TriangArray[t1].i != v1) &&  
          ( TriangArray[t1].j != v2)) && (flag != 3))           
  { ShiftVerticesRight(&TriangArray[t1]);   
    flag++;
  }

  if (flag < 3)
    Orientation = 1;  // left orientation ijk
  else
  { flag = 0;
    Orientation = 0;  // left orientation kij
    while ((( TriangArray[t1].k != v2) &&  
             ( TriangArray[t1].i != v1)) && (flag != 3))           
    { ShiftVerticesRight(&TriangArray[t1]);   
      flag++;
    }
  }

  long i, j, k, n_ij, n_jk, n_ki;
  if (Orientation) // triangle is ijk
  { i = TriangArray[t1].i;
    j = TriangArray[t1].j;
    k = TriangArray[t1].k;
    n_jk = TriangArray[t1].jk;
    n_ki = TriangArray[t1].ki;

    TriangArray[t1].j = nv;
    TriangArray[t1].jk = NextTriangPosition;
    
    TriangArray[NextTriangPosition].i = nv;
    TriangArray[NextTriangPosition].j = j;
    TriangArray[NextTriangPosition].k = k;
    TriangArray[NextTriangPosition].ij = NOINDEX;
    TriangArray[NextTriangPosition].jk = n_jk;
    TriangArray[NextTriangPosition].ki = t1;
    TriangArray[NextTriangPosition].Eij =
    TriangArray[NextTriangPosition].Ejk =
    TriangArray[NextTriangPosition].Eki = TRUE;
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;


//    SetNeighbour(OldT, k, j,NextTriangPosition);
    if (n_jk != NOINDEX)
       ReplaceNeighbour1(n_jk, t1, NextTriangPosition);

 
    AFLow->Insert(Vertices[nv].x,
                  Vertices[nv].y, nv, NextTriangPosition);
    AFLow->UpdateTIndex(Vertices[i].x, i, nv, t1);

    CheckDT(t1, n_ki);
    CheckDT(n_jk, NextTriangPosition);
  }
  else // orientation is kij
  { i = TriangArray[t1].i;
    j = TriangArray[t1].j;
    k = TriangArray[t1].k;

    n_ij = TriangArray[t1].ij;
    n_jk = TriangArray[t1].jk;

    TriangArray[t1].i = nv;
    TriangArray[t1].ij = NextTriangPosition;

    TriangArray[NextTriangPosition].i = i;
    TriangArray[NextTriangPosition].j = j;
    TriangArray[NextTriangPosition].k = nv;
    TriangArray[NextTriangPosition].ij = n_ij;
    TriangArray[NextTriangPosition].jk = t1;
    TriangArray[NextTriangPosition].ki = NOINDEX;
    TriangArray[NextTriangPosition].Eij =
    TriangArray[NextTriangPosition].Ejk =
    TriangArray[NextTriangPosition].Eki = TRUE;
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

    if (n_jk != NOINDEX)
       ReplaceNeighbour1(n_jk, t1, NextTriangPosition);

    AFLow->Insert(Vertices[nv].x,
                  Vertices[nv].y, nv, NextTriangPosition);
    AFLow->UpdateTIndex(Vertices[i].x, i, nv, t1);

    CheckDT(t1, n_jk);
    CheckDT(n_ij, NextTriangPosition);
  } 



  NextTriangPosition++;

  return 1;
}

int SolveMiddleCase(long i, DoubleLinkedList* Left,
                            DoubleLinkedList* Right)
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  long Pair = Left->TriangleIndex;
  InsertNewTriangle(i, Left, Right);

  if (AFLow->Insert(x,y,i,NextTriangPosition) == 0)
    return 0;

  CheckDT(Pair, NextTriangPosition);
  NextTriangPosition++;

  int MiddleFlag = 1;

  MakeTrianglesToTheRight(MiddleFlag, x, y, i, Right);

  MakeTrianglesToTheLeft(MiddleFlag, x, y, i, Left);

  return 1;
}

long SetTIndex(long T,  long v1, long v2)
// returns triangle wich have border edge v1, v2;
// at first it starts with triangle T
{ if (TriangArray[T].BorderTriangleFlag)
  { if ( ((TriangArray[T].i == v1) && (TriangArray[T].j == v2)) ||
         ((TriangArray[T].i == v2) && (TriangArray[T].j == v1)))
      return T;
    
    if ( ((TriangArray[T].j == v1) && (TriangArray[T].k == v2)) ||
         ((TriangArray[T].j == v2) && (TriangArray[T].k == v1)))
      return T;

    if ( ((TriangArray[T].k == v1) && (TriangArray[T].i == v2)) ||
         ((TriangArray[T].k == v2) && (TriangArray[T].i == v1)))
      return T;
  }

  T--;
  while (T >= 0)
  { if (TriangArray[T].BorderTriangleFlag)
    { if ( ((TriangArray[T].i == v1) && (TriangArray[T].j == v2)) ||
           ((TriangArray[T].i == v2) && (TriangArray[T].j == v1)))
        return T;
    
      if ( ((TriangArray[T].j == v1) && (TriangArray[T].k == v2)) ||
           ((TriangArray[T].j == v2) && (TriangArray[T].k == v1)))
        return T;

      if ( ((TriangArray[T].k == v1) && (TriangArray[T].i == v2)) ||
           ((TriangArray[T].k == v2) && (TriangArray[T].i == v1)))
        return T;
    }
    T = T-1;
  }

  return 0;
}

int SolveLeftCase(long i)
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  
  DoubleLinkedList* L1 = AFLow->ReturnLeft();
  DoubleLinkedList* L2 = AFLow->ReturnNextRight();
  long StoreVertex = L1->VertexNo;

  long StoreTIndex;

  if (SignedArea(x, y, L1->key, L1->y, L2->key, L2->y) > 0)
//  if (PositiveSignedArea(x, y, L1->key, L1->y, L2->key, L2->y) > 0)
  { InsertNewTriangle(i, L1, L2);
    long TInd = L1->TriangleIndex;
    AFLow->DeleteLeft();
    if (AFLow->AppendAtTheLeft(x,y,i,NextTriangPosition) == 0)
      return 0;
    CheckDT(TInd, NextTriangPosition);
    StoreTIndex = SetTIndex(NextTriangPosition, i, StoreVertex);

    NextTriangPosition++;

    MakeTrianglesToTheRight(0, x, y, i, L2);
  }
  else
  { StoreTIndex = NOINDEX; 
    if (AFLow->AppendAtTheLeft(x,y,i,NextTriangPosition) == 0)
      return 0;
  }


  DoubleLinkedList* AFWalker = L2 = ReturnDCLLeft();
  DoubleLinkedList* NextRightVertex;
  if ((NextRightVertex = L2->RPointer) != NULL)
  { 

    while ((NextRightVertex !=  NULL) && (SignedArea(x, y, 
                            NextRightVertex->key,
                            NextRightVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) > 0  ) 
/*    while ((NextRightVertex !=  NULL) && (PositiveSignedArea(x, y, 
                            NextRightVertex->key,
                            NextRightVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) > 0  ) */

    { // new triangles have to be added
      CloseNewTriangleRightDown(i, StoreTIndex, AFWalker, NextRightVertex);
      StoreTIndex = NextTriangPosition;
      NextTriangPosition++;

      DeleteFromLeft(AFWalker);

      AFWalker = NextRightVertex;
      NextRightVertex = NextRightVertex->RPointer;

    }   
  }
  AFHighAppendLeft(x, y, i, StoreTIndex);
  return 1;
}



int SolveRightCase(long i)
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  
  DoubleLinkedList* L1 = AFLow->ReturnRight();
  DoubleLinkedList* L2 = AFLow->ReturnNextLeft();
  long ti = L2->TriangleIndex;
  long StoreVertex = L1->VertexNo;

  long StoreTIndex;
  if (SignedArea(x, y, L1->key, L1->y, L2->key, L2->y) < 0)
//  if (NegativeSignedArea(x, y, L1->key, L1->y, L2->key, L2->y) < 0)
  { InsertNewTriangle(i, L2, L1);
    long TInd = L2->TriangleIndex;
    AFLow->DeleteRight();
    if (AFLow->AppendAtTheRight(x,y,i,NextTriangPosition) == 0)
      return 0;
    CheckDT(TInd, NextTriangPosition);
    StoreTIndex = SetTIndex(NextTriangPosition, i, StoreVertex);

    NextTriangPosition++;

    MakeTrianglesToTheLeft(0, x, y, i, L2);
  }
  else
  { StoreTIndex = NOINDEX;
    if (AFLow->AppendAtTheRight(x,y,i,NextTriangPosition) == 0)
      return 0;
  }


  DoubleLinkedList* AFWalker = L2 = ReturnDCLRight();
  DoubleLinkedList* NextLeftVertex;

  if ((NextLeftVertex = L2->LPointer) != NULL)
  { 

    while ((NextLeftVertex !=  NULL) && (SignedArea(x, y, 
                            NextLeftVertex->key,
                            NextLeftVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) < 0  )
/*  while ((NextLeftVertex !=  NULL) && (NegativeSignedArea(x, y, 
                            NextLeftVertex->key,
                            NextLeftVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) < 0  ) */
    { // new triangles have to be added
      CloseNewTriangleLeftDown(i, StoreTIndex, NextLeftVertex, AFWalker);
      StoreTIndex = NextTriangPosition;
      NextTriangPosition++;

      DeleteFromRight(AFWalker);

      AFWalker = NextLeftVertex;
      NextLeftVertex = NextLeftVertex->LPointer;
    }

  }
  AFHighAppendRight(x, y, i, StoreTIndex);
  return 1;
}


void  PlotTrianglesSweep(HDC hdc, double BoxXMin, double BoxYMin, 
                                  double BoxXMax, double BoxYMax, 
                                  int right, int bottom)
{ long ii;
  double x1, x2, x3, y1, y2, y3;
  int xi1, yi1, xi2, yi2, xi3, yi3;



  POINT plxy[5];

  HPEN hpen, hpenOld;
  HBRUSH HBrush = CreateSolidBrush(RGB(0,255,0));
  (HBRUSH)SelectObject(hdc, HBrush);

  hpen = CreatePen(PS_SOLID, 1, RGB(255,0,0));
  hpenOld = (HPEN)SelectObject(hdc, hpen);

  long i, j, k;
//  NextTriangPosition--;
  for (ii = 0; ii < NextTriangPosition; ii++)
  { i = TriangArray[ii].i;
    j = TriangArray[ii].j;
    k = TriangArray[ii].k;

    x1 = Vertices[i].x;
    y1 = Vertices[i].y;

    x2 = Vertices[j].x;
    y2 = Vertices[j].y;

    x3 = Vertices[k].x;
    y3 = Vertices[k].y;

    xi1 = FromRealToPixelX(x1, BoxXMin, BoxXMax - BoxXMin, right);
    yi1 = bottom - FromRealToPixelY(y1, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi2 = FromRealToPixelX(x2, BoxXMin, BoxXMax - BoxXMin, right);
    yi2 = bottom - FromRealToPixelY(y2, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi3 = FromRealToPixelX(x3, BoxXMin, BoxXMax - BoxXMin, right);
    yi3 = bottom - FromRealToPixelY(y3, BoxYMin, BoxYMax - BoxYMin, bottom);

    plxy[0].x = xi1;   plxy[0].y = yi1;
    plxy[1].x = xi2;   plxy[1].y = yi2;
    plxy[2].x = xi3;   plxy[2].y = yi3;

//    Polygon(hdc, plxy, 3);
 
    MoveToEx(hdc, xi1, yi1, NULL);
    LineTo(hdc, xi2, yi2);
    LineTo(hdc, xi3, yi3);
    LineTo(hdc, xi1, yi1);
  }

  SelectObject(hdc, hpenOld);
  DeleteObject(hpen);
}

int PrepareTriangles(long t1, long t2)
{
  while (( TriangArray[t1].k == TriangArray[t2].i) || 
         ( TriangArray[t1].k == TriangArray[t2].j) || 
         ( TriangArray[t1].k == TriangArray[t2].k))
    ShiftVerticesRight(&TriangArray[t1]);  


  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangArray[t2]);

  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangArray[t2]);
  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;
  return 0;
}


void SHRVertices(BaseTriangle* t)
// shift vertices right
{ long a;

  a = t->k;
  t->k = t->j;
  t->j = t->i;
  t->i = a;

  a = t->ki;
  t->ki = t->jk;
  t->jk = t->ij;
  t->ij = a;
}


void ReplaceNeighbour1(long TIndex, long OldT, long NewNeighbour)
{ if (TriangArray[TIndex].ij == OldT) 
    TriangArray[TIndex].ij = NewNeighbour;
  else
  { if (TriangArray[TIndex].jk == OldT)
      TriangArray[TIndex].jk = NewNeighbour;
    else
    { if (TriangArray[TIndex].ki == OldT)
       TriangArray[TIndex].ki = NewNeighbour;
    }
  }
}


void CheckDT(long t1, long t2)
// Triangles t1 and t2 are checked for Delaunay criterium and 
// legalised in neccessary
{ 

  if ((t1 != NOINDEX) && (t2 != NOINDEX))
   if (t1 != t2)
    { long i1, j1, k1, k2;

      if (PrepareTriangles(t1, t2))
      {
        if (PrepareTriangles(t2, t1))
        {


          i1 = TriangArray[t1].i;
          j1 = TriangArray[t1].j;
          k1 = TriangArray[t1].k; 
          k2 = TriangArray[t2].k;     
          C[0][0] = Vertices[i1].x - Vertices[k2].x;
          C[0][1] = Vertices[i1].y - Vertices[k2].y;
          C[0][2] = C[0][0] * C[0][0] + C[0][1] * C[0][1];

          C[1][0] = Vertices[j1].x - Vertices[k2].x;
          C[1][1] = Vertices[j1].y - Vertices[k2].y;
          C[1][2] = C[1][0] * C[1][0] + C[1][1] * C[1][1];

          C[2][0] = Vertices[k1].x - Vertices[k2].x;
          C[2][1] = Vertices[k1].y - Vertices[k2].y;
          C[2][2] = C[2][0] * C[2][0] + C[2][1] * C[2][1];

          double D =   C[0][0] * (C[1][1] * C[2][2] - C[1][2] * C[2][1])
                     + C[0][1] * (C[1][2] * C[2][0] - C[1][0] * C[2][2])
                     + C[0][2] * (C[1][0] * C[2][1] - C[1][1] * C[2][0]);

 
          if(D < 0) 
          { 

            long n_ij1 = TriangArray[t1].ij; 
            long n_jk1 = TriangArray[t1].jk;
            long n_ki1 = TriangArray[t1].ki;

            long n_ij2 = TriangArray[t2].ij; 
            long n_jk2 = TriangArray[t2].jk;
            long n_ki2 = TriangArray[t2].ki;

            // new triangles are at the location of the old triangles
            TriangArray[t1].i = k1;     
            TriangArray[t1].j = i1;   
            TriangArray[t1].k = k2;
      
            TriangArray[t2].i = k1;     
            TriangArray[t2].j = k2;   
            TriangArray[t2].k = j1;


            // store triangles surrounding the point
//            DeleteTriangleFromPoint(i1, t2);  
//            DeleteTriangleFromPoint(j1, t1);  
//            AddTriangleToPoint(k1, t2);
//            AddTriangleToPoint(k2, t1);

            // update the neighbours
            TriangArray[t1].ij = n_ki1;
            TriangArray[t1].jk = n_jk2;
            TriangArray[t1].ki = t2;

            TriangArray[t2].ij = t1;
            TriangArray[t2].jk = n_ki2;
            TriangArray[t2].ki = n_jk1;

            if (n_jk2 != NOINDEX)
              ReplaceNeighbour1(n_jk2, t2, t1);      

            if (n_jk1 != NOINDEX)
              ReplaceNeighbour1(n_jk1, t1, t2);
 
            CheckBorderFlag(t1);
            CheckBorderFlag(t2);
            if (TriangArray[t1].BorderTriangleFlag)
              UpdateTriangleIndex(t1);

              if (TriangArray[t2].BorderTriangleFlag)
                UpdateTriangleIndex(t2);

              CheckBorderFlag(t1);
              CheckBorderFlag(t2);
  
/*            if (BorderFlagT1)
              AFLow->UpdateTIndex(t1, 
                       Vertices[TriangArray[t1].i],
                       Vertices[TriangArray[t1].j],
                       Vertices[TriangArray[t1].i]);

            if (BorderFlagT2)
              AFLow->UpdateTIndex(t2, 
                       Vertices[TriangArray[t2].i],
                       Vertices[TriangArray[t2].j],
                       Vertices[TriangArray[t2].i]);
*/
            CheckDT(t1, n_jk2);
            CheckDT(t2, n_ki2);
            CheckDT(t1, n_ki1);
            CheckDT(t2, n_jk1);

          }

        }
      }
    }
  

}

void UpdateTriangleIndex(int ti)

// ta funkcija je napacna. Azuriranje AFLow je dovoljeno le glede na rob;
// prav tako je smiselno primerjati le indekse oglisc.
{ int DoneFlag = 0;

  if (TriangArray[ti].ij == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].i].x,
                                   TriangArray[ti].i, 
                                   TriangArray[ti].j, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].i, TriangArray[ti].j, ti);
  }

  if (TriangArray[ti].jk == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].j].x,
                                   TriangArray[ti].j, 
                                   TriangArray[ti].k, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].j, TriangArray[ti].k, ti);
  }

  if (TriangArray[ti].ki == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].k].x,
                                   TriangArray[ti].k, 
                                   TriangArray[ti].i, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].k, TriangArray[ti].i, ti);
  }
}


void PrintTriangulation()
{ int i, bf;

  FILE* f = fopen("TRIANGLES.TXT", "wt");
  for (i = 0; i < NextTriangPosition; i++)
  { if (TriangArray[i].BorderTriangleFlag) bf = 1;
    else bf = 0;

    fprintf(f, "%d. i:%ld j:%ld k:%ld ij:%ld jk:%ld ki:%ld BTF:%d\n",
                i, TriangArray[i].i, TriangArray[i].j, TriangArray[i].k, 
               TriangArray[i].ij, TriangArray[i].jk, TriangArray[i].ki,bf);
  }

  fclose(f);
}

//  PlotTriangulatedPoSweep, ints(hdc, BoxXMin, BoxYMin, BoxXMax, BoxYMax, r.right, r.bottom, NoOfPoints);



double PositiveSignedArea(double x, double y, double x1, double y1,
                          double x2, double y2)
{
  if (y2 >= y1)
    return 1;
  return SignedArea(x, y, x1, y1, x2, y2);
}
   

double NegativeSignedArea(double x, double y, double x1, double y1,
                          double x2, double y2)
{
  if (y1 >= y2)
    return -1;
  return SignedArea(x, y, x1, y1, x2, y2);
}                         
