// USPTriangF.cpp: implementation of the USPTriangF class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "USPTriangF.h"
#include "Proximity2LNew.h"
#include <stdio.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// global variables

static double BoxXMin, BoxYMin, BoxXMax, BoxYMax;  // bounding box
static MyPointT* Points = NULL;                    // points to be triangulated
static Proximity2L* Proximity = NULL;              // pointer to 2LUP searching struct 
static long NoOfVertices = 0;
static long OriginalNoOfVertices = 0;
static BaseTriangle* TriangleArray = NULL;
static long NoOfLegalisations= 0;

long NextTrianglePosition;      
long CurrentPointIndex;     
double MyEpsilon;
double M[3][3];

static long PointInTriangleCounter = 0;
static long TheNextClosestPointCounter = 0;

long ReturnPointInTriangleCounter()
{ return PointInTriangleCounter; }

long ReturnTheNextClosestPointCounter()
{ return TheNextClosestPointCounter; }


// samo zacasno za testiranje sosednosti

int DvaIndeksaSkupna(long a, long b, long tp)
{ int counter = 0;

  if ((a == TriangleArray[tp].i) ||
      (a == TriangleArray[tp].j) ||
      (a == TriangleArray[tp].k) )
    counter++;


  if ((b == TriangleArray[tp].i) ||
      (b == TriangleArray[tp].j) ||
      (b == TriangleArray[tp].k) )
    counter++;

  return counter;
}

int SosediPravi()
{ long u;
  long sosed;
  int res;

  for (u = 0; u < NextTrianglePosition; u++)
  { sosed = TriangleArray[u].ij;
    
    if (sosed != NONE)
      if ((res = DvaIndeksaSkupna(TriangleArray[u].i, TriangleArray[u].j, sosed)) != 2)
        return res;


    sosed = TriangleArray[u].jk;
    if (sosed != NONE)
      if ((res = DvaIndeksaSkupna(TriangleArray[u].j, TriangleArray[u].k, sosed)) == 0)
        return res;

    sosed = TriangleArray[u].ki;

    if (sosed != NONE)
      if ((res = DvaIndeksaSkupna(TriangleArray[u].k, TriangleArray[u].i, sosed)) == 0)
        return res;
  }
return 2;

}





void InitUSPTriang(long n, double *XArr, double *YArr,
                   double BXMin, double BYMin, 
                   double BXMax, double BYMax)
{  MyEpsilon = 0.000000000001;

  NoOfLegalisations= 0;

  BoxXMin =  BXMin;
  BoxYMin =  BYMin;
  BoxXMax =  BXMax;
  BoxYMax =  BYMax;

  double aaa, bbb;
  aaa = ABS(BoxXMax - BoxXMin);
  bbb = ABS(BoxYMax - BoxYMin);
  if (aaa > bbb) MyEpsilon *= bbb;
  else MyEpsilon *= aaa;

  

 
  OriginalNoOfVertices = n;
  NoOfVertices = n+3;
  TriangleArray = new BaseTriangle[2*NoOfVertices]; 

  Points = new MyPointT[OriginalNoOfVertices + 10];

  int jj;
  for (long j = 0; j < NoOfVertices; j++)
  { 

    Points[j].x = XArr[j]; 
    Points[j].y = YArr[j]; 
    Points[j].Index = j;  
    Points[j].TriangulatedFlag = true;

    Points[j].ListOfTriangles.ListofTriangles = NULL;
    Points[j].ListOfTriangles.NoT = 0;
    for (jj = 0; jj < NO_OF_SURROUNDING_TRIANGLES; jj++)
      Points[j].ListOfTriangles.FlagsArray[jj] = false;
  }


//  Proximity = new Proximity2L(BoxXMin, BoxYMin, BoxXMax, BoxYMax, n);

  if (InitProximityStructure(Points, BoxXMin, BoxYMin, BoxXMax, BoxYMax, n) == 0)
  {
    int skociven = 0;
  }



  double a,b;
  if (ABS(BoxXMin) > ABS(BoxXMax)) a = BoxXMin;
  else a = BoxXMax;

  if (ABS(BoxYMin) > ABS(BoxYMax)) b = BoxYMin;
  else b = BoxYMax;

  if (b > a) a = b;

   a *= 10;
  // set big triangle
  long i = n;
  Points[i].x = a;  Points[i].y = 0;  Points[i].Index = i++;
  Points[i].x = -a; Points[i].y = -a; Points[i].Index = i++;
  Points[i].x = 0;  Points[i].y = a;  Points[i].Index = n;

  NextTrianglePosition = 3;  
}


void FinalUSPTriang()
{
  ReleaseProximityStructure();

//  if (Proximity != NULL)
//    delete Proximity;

//  Proximity = NULL;

  if (TriangleArray != NULL)
    delete[] TriangleArray;

  TriangleArray = NULL;

  SurroundingTriangleInfo* pa;
  ListOfTrianglesType* t;
  for (long i = 0; i < NoOfVertices; i++)
  { pa = &Points[i].ListOfTriangles;
    while (pa->ListofTriangles != NULL)
    { t = pa->ListofTriangles;
      pa->ListofTriangles = pa->ListofTriangles->next;
      delete t;
    }
  }
  delete[] Points;
  Points = NULL;
}



void InsertTheFirstPoint()
{ long i = OriginalNoOfVertices;
  long j = OriginalNoOfVertices + 1;
  long k = OriginalNoOfVertices + 2;
  long L = 0;

  TriangleArray[0].i = i;     TriangleArray[0].j = j;   TriangleArray[0].k = L;
  TriangleArray[0].ij = NONE; TriangleArray[0].jk = 1;  TriangleArray[0].ki = 2;

  TriangleArray[1].i = j;     TriangleArray[1].j = k;   TriangleArray[1].k = L;
  TriangleArray[1].ij = NONE; TriangleArray[1].jk = 2;  TriangleArray[1].ki = 0;

  TriangleArray[2].i = k;     TriangleArray[2].j = i;   TriangleArray[2].k = L;
  TriangleArray[2].ij = NONE; TriangleArray[2].jk = 0;  TriangleArray[2].ki = 1;


  // store triangles surrounding the point
  AddTriangleToPoint(0, 0);
  AddTriangleToPoint(0, 1);
  AddTriangleToPoint(0, 2);

  NextTrianglePosition = 3; // points always to the first free location in VertexArray
//  Proximity->InsertPoint(Points[0].x, Points[0].y, 0);

  InsertPoint2USP(0);
}

void AddTriangleToPoint(long index, long t1)
// function inserts the index of a triangle t1 to the list of surrounding
// triangles of vertex index. this list combines statical and dinamical data 
// structure.
{ if (index < OriginalNoOfVertices)  // do not add to the artificial vertices
  { SurroundingTriangleInfo* pa = &Points[index].ListOfTriangles;

    if (pa->NoT < NO_OF_SURROUNDING_TRIANGLES)
    { // add to static array
      int i = 0;
      bool flag = true;
      while ((i < NO_OF_SURROUNDING_TRIANGLES) && (flag))
        if (pa->FlagsArray[i] == false)
        { pa->SurroundingArray[i] = t1;
          flag = false;
          pa->FlagsArray[i] = true;
        }
        else
          i++;
    } 
    else //add to the dinamic list;
    { ListOfTrianglesType* tp = new ListOfTrianglesType;
      tp->TriangleIndex = t1;
      tp->next = pa->ListofTriangles;
      pa->ListofTriangles = tp;
    }
    pa->NoT++;
  }
}


void DeleteTriangleFromPoint(long index, long t1)
// function deletess the index of a triangle t1 from the list of surrounding
// triangles of vertex index. this list combines statical and dinamical data 
// structure.
{ 
  if (index < OriginalNoOfVertices)  
  { SurroundingTriangleInfo* pa = &Points[index].ListOfTriangles;
 
    int i = 0;
    bool flag = true;    

    while ((i < NO_OF_SURROUNDING_TRIANGLES) && flag)
    { if (pa->FlagsArray[i])
      { if (pa->SurroundingArray[i] == t1)
        { pa->FlagsArray[i] = false;
          flag = false;
          pa->NoT--;

          ListOfTrianglesType* head;

          if ((head = pa->ListofTriangles) != NULL) // list is not empty
          { pa->ListofTriangles = head->next;
            pa->FlagsArray[i] = true;
            pa->SurroundingArray[i] = head->TriangleIndex;
            delete head;
          }
        }
      } 
      i++;
    }

    if (flag)  // triangle is not found in the array, let us check the list
    { ListOfTrianglesType* head = pa->ListofTriangles;      
      ListOfTrianglesType* tail = head;

      while ((head != NULL) && (!flag))
      { if (head->TriangleIndex == t1)
        { flag = true;
          pa->NoT--;
          if (head == tail) // first record
          { pa->ListofTriangles = head->next;
            delete head;
          }
          else
          { tail->next = head->next;
            delete head;
          }
        } 
        else
        { tail = head;
          head = head->next;
        }
      } 
    }
  }
}



long ReturnTriangle(long Position)
{

  return NO_MORE_TRIANGLES;
}



long FindTheClosestPoint()
{ double x = Points[CurrentPointIndex].x; 
  double y = Points[CurrentPointIndex].y;
  long TheClosestIndex;

  if (CurrentPointIndex < START_2USP) // use brute force
  { long i;
    long storeI = 0;
    double d = MAX_DOUBLE;
    double d1;
    double a, b;

    for (i = 0; i < CurrentPointIndex; i++)
    { Points[i].CloseCheckFlag = 0;
      if (Equal(x, Points[i].x, MyEpsilon) && Equal(y, Points[i].y, MyEpsilon) )
        return THE_SAME_POINT;
      else
      { a = Points[i].x - x;
        b = Points[i].y - y;
        d1 = a * a + b * b;
//        d1 = Sqr(Points[i].x - x) + Sqr (Points[i].y - y);
        if ( d1 < d)
        { storeI = i;
          d = d1;
        }
      }
    }
    Points[storeI].CloseCheckFlag = 1;
    return storeI;
  }
  else // use 2USP
  {
/*   Proximity->InitTheClosePoint(x, y);
    bool TheSamePoints;
    TheClosestIndex = Proximity->DetermineTheClosestPoint(TheSamePoints);
*/
    TheClosestIndex = DetermineTheClosestPoint(CurrentPointIndex);

//    if (TheSamePoints == true)
//      return THE_SAME_POINT;
    return TheClosestIndex;
  }
}


long FindTheNextClosestPoint(long Previous)
{

TheNextClosestPointCounter++;

  if (CurrentPointIndex < START_2USP)
  { double x = Points[CurrentPointIndex].x; 
    double y = Points[CurrentPointIndex].y;
    
    double d1 = MAX_DOUBLE;
    long storeI = NONE;
    double d; 
    long i;
    double a, b;
  

    for (i = 0; i < CurrentPointIndex; i++)
    { if (Points[i].CloseCheckFlag == 0)
      { a = Points[i].x - x;
        b = Points[i].y - y;
        d = a * a + b * b; 
//d = Sqr(Points[i].x - x) + Sqr (Points[i].y - y);
        if (d  < d1)
        { storeI = i;
          d1 = d;
        }
      }
    }  
    if (storeI != NONE)
      Points[storeI].CloseCheckFlag  = 1;
    else
    { int qqq = 0;} 
    return storeI;
  }

//  bool TheSamePoints;
//  return Proximity->DetermineTheClosestPoint(TheSamePoints);

  return ReturnTheNextClosestPoint();
}

bool PointOnEdge(double x1, double y1, double x2, double y2, 
                 double tpx, double tpy, double& a1)
// Method determines if point (txp, typ) lies on the edge determined by
// end points (x1, y1) (x2, y2). Cross product is calculated for that
// and the result of the cross product is returned in a1. 
// If point is on the line determined by (x1, y1) (x2, y2),cross product is 
// 0. Then, we have to determined if point is really on the edge. If it is
// method returns TRUE, otherwise, it returns FALSE.
{ 
  // calculate signed area
/*
  double dx1 = tpx - x1;
  double dy1 = tpy - y1;
  double dx2 = x2 - x1;
  double dy2 = y2 - y1;

  if (Equal(dx1*dy2, dy1 * dx2, MyEpsilon))
    return true;

  a1 = (x1 * (y2 - tpy) - y1 * (x2 - tpx) + (x2 * tpy - y2 * tpx));
  return false;
*/

  a1 = (x1 * (y2 - tpy) - y1 * (x2 - tpx) + (x2 * tpy - y2 * tpx));

  if (Equal(a1, 0, MyEpsilon))
  { 
    if (x1 > x2) Swap(x1, x2);
    if (y1 > y2) Swap(y1, y2);

    if (Equal(x1, x2, MyEpsilon)) // vertical line segment
    { if ((tpy > y1) && (tpy < y2))
        return true;
      else
        return false;
    }

    if (Equal(y1, y2, MyEpsilon)) // horizontal line segment
    { if ((tpx > x1)  && (tpx < x2))
        return true;
      else
        return false; 
    }

    if ((tpx > x1) && (tpx < x2) && (tpy > y1) && (tpy < y2))
      return true;
  }
  
  return false;

}


int ExamineTriangle(double tpx, double tpy, long index, long &t1, long &t2)
{   
  t1 = t2 = NONE;

PointInTriangleCounter++;

  double x1 = Points[TriangleArray[index].i].x;
  double y1 = Points[TriangleArray[index].i].y;
  double x2 = Points[TriangleArray[index].j].x;
  double y2 = Points[TriangleArray[index].j].y;
  double x3 = Points[TriangleArray[index].k].x;
  double y3 = Points[TriangleArray[index].k].y;

  bool EdgeFlag1, EdgeFlag2, EdgeFlag3;
  double a1, a2, a3;

       
   EdgeFlag1 = PointOnEdge(x1, y1, x2, y2, tpx, tpy, a1);
   if (EdgeFlag1)
   { t1 = index;
     t2 = TriangleArray[index].ij;
     return 101; // point is on edge ti-tj
   }  

   if (a1 > 0)  return 0;

   EdgeFlag2 = PointOnEdge(x2, y2, x3, y3, tpx, tpy, a2);
   if (EdgeFlag2)
   { t1 = index;
     t2 = TriangleArray[index].jk;
     return 102; // point is on edge tj-tk    
   }

   if (a2 > 0)  return 0;
    
   EdgeFlag3 = PointOnEdge(x3, y3, x1, y1, tpx, tpy,a3);
   if (EdgeFlag3)
   { t1 = index;
     t2 = TriangleArray[index].ki; 
     return 103; // point is on edge tk-ti
   }

   if (a3 > 0) return 0;

   t1 =  index;   
   return 1;
}

int PointInTriangle(long TheClosestIndex, long& t1, long& t2)
// Function determines if point CurrentPointIndex is inside any of trinagle 
// surrounding the the point TheClosestIndex
// it returns:
// - t1: index of the triangle into which point p exist
// - t2: index of the second tringle, if point is on the triangle edge
// these events are returned also as the following flags
// 0: point is outside
// 1: point is inside
// 101: point is on the edge ti-tj
// 102: point is on the edge tj-tk
// 103: point is on the edge tk-ti
// 200: point is the same as one of the point inserted already 
{ long i = 0;

  
  long* IndexArray = Points[TheClosestIndex].ListOfTriangles.SurroundingArray; 
  bool* FlagArray = Points[TheClosestIndex].ListOfTriangles.FlagsArray; 
  double  tpx = Points[CurrentPointIndex].x;
  double  tpy = Points[CurrentPointIndex].y;

  int code;



  while (i < NO_OF_SURROUNDING_TRIANGLES) // spremenil 30.6.2001
//  while (i < Points[TheClosestIndex].ListOfTriangles.NoT)
  { if (FlagArray[i])
    { if ( (code = ExamineTriangle(tpx, tpy, IndexArray[i], t1, t2)) != 0)
        return code;
    } 
    i++;
  }

  ListOfTrianglesType* list = Points[CurrentPointIndex].ListOfTriangles.ListofTriangles;

  while (list != NULL)
    if ( (code = ExamineTriangle(tpx, tpy, IndexArray[i], t1, t2)) != 0)
      return code;
    else 
      list = list->next;
  
  return NONE;
}


void ReplaceNeighbour(long TIndex, long OldT, long NewNeighbour)
{ if (TriangleArray[TIndex].ij == OldT) 
    TriangleArray[TIndex].ij = NewNeighbour;
  else
  { if (TriangleArray[TIndex].jk == OldT)
      TriangleArray[TIndex].jk = NewNeighbour;
    else
    { if (TriangleArray[TIndex].ki == OldT)
       TriangleArray[TIndex].ki = NewNeighbour;
    }
  }
}


void ReplaceNeighbours(long TIndex, long OldT1, long NewT1, 
                                   long OldT2, long NewT2)
{ bool Eij = true;
  bool Ejk = true;
  bool Eki = true;


  if (TriangleArray[TIndex].ij == OldT1) 
  { TriangleArray[TIndex].ij = NewT1;
    Eij = false;
  }
  else
  { if (TriangleArray[TIndex].jk == OldT1)
    { TriangleArray[TIndex].jk = NewT1;
      Ejk = false;
    }
    else
    { if (TriangleArray[TIndex].ki == OldT1)
      { TriangleArray[TIndex].ki = NewT1;
        Eki = false;
      }
    }
  }


  if ((TriangleArray[TIndex].ij == OldT2) && Eij)
    TriangleArray[TIndex].ij = NewT2;      
  else
  { if ((TriangleArray[TIndex].jk == OldT2) && Ejk)
      TriangleArray[TIndex].jk = NewT2;
    else
    { if ((TriangleArray[TIndex].ki == OldT2) && Eki)
      { TriangleArray[TIndex].ki = NewT2;        
      }
    }
  }
}



void SplitOneTriangle(long OldT)
{ // set old neighbours
  long n_ij = TriangleArray[OldT].ij;
  long n_jk = TriangleArray[OldT].jk;
  long n_ki = TriangleArray[OldT].ki;

   // old vertices
  long i = TriangleArray[OldT].i;
  long j = TriangleArray[OldT].j;
  long k = TriangleArray[OldT].k;
  long L = CurrentPointIndex;

  // create new three triangles
  long a1 = OldT; // the old triangles is overwritten
  long b1 = NextTrianglePosition++;
  long c1 = NextTrianglePosition++;
  TriangleArray[a1].i = i;     
  TriangleArray[a1].j = j;   
  TriangleArray[a1].k = L;

  TriangleArray[b1].i = j;     
  TriangleArray[b1].j = k;   
  TriangleArray[b1].k = L;

  TriangleArray[c1].i = k;     
  TriangleArray[c1].j = i;   
  TriangleArray[c1].k = L;

  // store triangles surrounding the point
  AddTriangleToPoint(i, c1);

  AddTriangleToPoint(j, b1);

  DeleteTriangleFromPoint(k, a1);
  AddTriangleToPoint(k, b1);
  AddTriangleToPoint(k, c1);

  AddTriangleToPoint(L, a1);
  AddTriangleToPoint(L, b1);
  AddTriangleToPoint(L, c1);

  // update the neighbours
  TriangleArray[a1].ij = n_ij;
  TriangleArray[a1].jk = b1;
  TriangleArray[a1].ki = c1;

  TriangleArray[b1].ij = n_jk;
 // TriangleArray[b1].jk = a1;
 // TriangleArray[b1].ki = c1;
  TriangleArray[b1].jk = c1;
  TriangleArray[b1].ki = a1;

  TriangleArray[c1].ij = n_ki;
  TriangleArray[c1].jk = a1;
  TriangleArray[c1].ki = b1;  

  if (n_ij != NONE)
   ReplaceNeighbour(n_ij, OldT, a1);

  if (n_jk != NONE)
   ReplaceNeighbour(n_jk, OldT, b1);

  if (n_ki != NONE)
    ReplaceNeighbour(n_ki, OldT, c1);

  Legalise(a1, n_ij);
  Legalise(b1, n_jk);
  Legalise(c1, n_ki);
}

void SplitTwoTriangles(long t1, long t2)
{ if (t1 != t2)
  { long L = CurrentPointIndex;

    if (AdjustVertices(t1, t2))
    { if (AdjustVertices(t2, t1) )
      {
        // old vertices
        long i1 = TriangleArray[t1].i;
        long j1 = TriangleArray[t1].j;
        long k1 = TriangleArray[t1].k;
        long i2 = TriangleArray[t2].i;
        long j2 = TriangleArray[t2].j;
        long k2 = TriangleArray[t2].k;

        long ki1 = TriangleArray[t1].ki;
        long jk1 = TriangleArray[t1].jk;
        long jk2 = TriangleArray[t2].jk;
        long ki2 = TriangleArray[t2].ki;

        long a1 = t1;
        long b1 = t2;
        TriangleArray[a1].i = i1;     
        TriangleArray[a1].j = L;   
        TriangleArray[a1].k = k1;

        TriangleArray[b1].i = j1;     
        TriangleArray[b1].j = k1;   
        TriangleArray[b1].k = L;

        long c1 = NextTrianglePosition++;
        TriangleArray[c1].i = k2;     
        TriangleArray[c1].j = L;   
        TriangleArray[c1].k = i1;
  
        long d1 = NextTrianglePosition++;
        TriangleArray[d1].i = j1;     
        TriangleArray[d1].j = L;   
        TriangleArray[d1].k = k2;

        // store triangles surrounding the point  
        DeleteTriangleFromPoint(i1, b1);
        AddTriangleToPoint(i1, c1);

        DeleteTriangleFromPoint(j1, a1);
        AddTriangleToPoint(j1, b1);
        AddTriangleToPoint(j1, d1);

        AddTriangleToPoint(k1, b1);

        DeleteTriangleFromPoint(k2, b1);
        DeleteTriangleFromPoint(k2, a1); 
        AddTriangleToPoint(k2, d1);
        AddTriangleToPoint(k2, c1);

        AddTriangleToPoint(L, a1);
        AddTriangleToPoint(L, b1);
        AddTriangleToPoint(L, c1);
        AddTriangleToPoint(L, d1);

        // update the neighbours
        TriangleArray[a1].ij = c1;  
        TriangleArray[a1].jk = b1;
        TriangleArray[a1].ki = ki1;

        TriangleArray[b1].ij = jk1;  
        TriangleArray[b1].jk = a1;
        TriangleArray[b1].ki = d1;

        TriangleArray[c1].ij = d1;  
        TriangleArray[c1].jk = a1;
        TriangleArray[c1].ki = jk2;

        TriangleArray[d1].ij = b1;  
        TriangleArray[d1].jk = c1;
        TriangleArray[d1].ki = ki2;


        if ((jk1 == ki2) && (jk1 != NONE))
                        // the triangles form the concave angle
                        // through edges j1k1 and k2i2
        {  ReplaceNeighbours(jk1, a1, b1, b1, d1);

            if (jk2 != NONE)
              ReplaceNeighbour(jk2, b1, c1);

        }
        else
        { if ((ki1 == jk2) && (ki1 != NONE) ) 
          { ReplaceNeighbours(ki1, a1, a1, b1, c1);

            if (jk1 != NONE)
              ReplaceNeighbour(jk1, a1, b1);

            if (ki2 != NONE)
              ReplaceNeighbour(ki2, b1, d1);
          }
          else
          {
            if (jk1 != NONE)
              ReplaceNeighbour(jk1, a1, b1);

            if (jk2 != NONE)
              ReplaceNeighbour(jk2, b1, c1);

            if (ki2 != NONE)
              ReplaceNeighbour(ki2, b1, d1);
          }
        } 


/*

        if (jk1 != NONE)
          ReplaceNeighbour(jk1, a1, b1);

        if (jk2 != NONE)
          ReplaceNeighbour(jk2, b1, c1);

        if (ki2 != NONE)
          ReplaceNeighbour(ki2, b1, d1);
*/


  //  if (ki1 != NONE)                    not neccessary
//    ReplaceNeighbour(ki1, a1, a1);

        Legalise(a1, ki1);
        Legalise(b1, jk1);
        Legalise(c1, jk2);
        Legalise(d1, ki2); 
      }
    }  
  }
}


int AdjustVertices(long t1, long t2)
{ 
  while (( TriangleArray[t1].k == TriangleArray[t2].i) || 
         ( TriangleArray[t1].k == TriangleArray[t2].j) || 
         ( TriangleArray[t1].k == TriangleArray[t2].k))
    ShiftVerticesRight(&TriangleArray[t1]);  


  if ((TriangleArray[t1].i == TriangleArray[t2].j) && 
      (TriangleArray[t1].j == TriangleArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangleArray[t2]);

  if ((TriangleArray[t1].i == TriangleArray[t2].j) && 
      (TriangleArray[t1].j == TriangleArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangleArray[t2]);
  if ((TriangleArray[t1].i == TriangleArray[t2].j) && 
      (TriangleArray[t1].j == TriangleArray[t2].i))
    return 1;
  return 0;
}


void ShiftVerticesRight(BaseTriangle* t)
{ long a;

  a = t->k;
  t->k = t->j;
  t->j = t->i;
  t->i = a;

  a = t->ki;
  t->ki = t->jk;
  t->jk = t->ij;
  t->ij = a;
}


void Legalise(long t1, long t2)
// Triangles t1 and t2 are checked for Delaunay criterium and 
// legalised in neccessary
{ if ((t1 != NONE) && (t2 != NONE))
  { if (t1 != t2)
    { long i1, j1, k1, k2;

      if (AdjustVertices(t1, t2))
      {
        if (AdjustVertices(t2, t1))
        {

          i1 = TriangleArray[t1].i;
          j1 = TriangleArray[t1].j;
          k1 = TriangleArray[t1].k; 
          k2 = TriangleArray[t2].k;     

          M[0][0] = Points[i1].x - Points[k2].x;
          M[0][1] = Points[i1].y - Points[k2].y;
          M[0][2] = M[0][0] * M[0][0] + M[0][1] * M[0][1];

          M[1][0] = Points[j1].x - Points[k2].x;
          M[1][1] = Points[j1].y - Points[k2].y;
          M[1][2] = M[1][0] * M[1][0] + M[1][1] * M[1][1];

          M[2][0] = Points[k1].x - Points[k2].x;
          M[2][1] = Points[k1].y - Points[k2].y;
          M[2][2] = M[2][0] * M[2][0] + M[2][1] * M[2][1];

          double D =   M[0][0] * (M[1][1] * M[2][2] - M[1][2] * M[2][1])
                     + M[0][1] * (M[1][2] * M[2][0] - M[1][0] * M[2][2])
                     + M[0][2] * (M[1][0] * M[2][1] - M[1][1] * M[2][0]);

          if(D < 0) 
          { NoOfLegalisations++;
            long n_ij1 = TriangleArray[t1].ij; 
            long n_jk1 = TriangleArray[t1].jk;
            long n_ki1 = TriangleArray[t1].ki;

            long n_ij2 = TriangleArray[t2].ij; 
            long n_jk2 = TriangleArray[t2].jk;
            long n_ki2 = TriangleArray[t2].ki;

            // new triangles are at the location of the old triangles
            TriangleArray[t1].i = k1;     
            TriangleArray[t1].j = i1;   
            TriangleArray[t1].k = k2;
      
            TriangleArray[t2].i = k1;     
            TriangleArray[t2].j = k2;   
            TriangleArray[t2].k = j1;


            // store triangles surrounding the point
            DeleteTriangleFromPoint(i1, t2);  
            DeleteTriangleFromPoint(j1, t1);  
            AddTriangleToPoint(k1, t2);
            AddTriangleToPoint(k2, t1);

            // update the neighbours
            TriangleArray[t1].ij = n_ki1;
            TriangleArray[t1].jk = n_jk2;
            TriangleArray[t1].ki = t2;

            TriangleArray[t2].ij = t1;
            TriangleArray[t2].jk = n_ki2;
            TriangleArray[t2].ki = n_jk1;

            if (n_jk2 != NONE)
              ReplaceNeighbour(n_jk2, t2, t1);      

            if (n_jk1 != NONE)
              ReplaceNeighbour(n_jk1, t1, t2);
 
            Legalise(t1, n_jk2);
            Legalise(t2, n_ki2);
          }
        }
      }
    }
  }
}


void InsertPoint(long ci)
{ CurrentPointIndex = ci;

  long TheClosestIndex = FindTheClosestPoint();

 
  if (TheClosestIndex != THE_SAME_POINT)
  {
    bool found = false; 
    long FoundTriangle1, FoundTriangle2;

    int EscapeFlag = 0;

    while (!found)
    { 
      
      PointInTriangle(TheClosestIndex, FoundTriangle1, FoundTriangle2);

      EscapeFlag++;
      if (EscapeFlag > 200)
      { 
        Points[CurrentPointIndex].TriangulatedFlag = false;
        goto ven;
      }

      if (FoundTriangle1 != NONE)
      { found = true;

        InsertPoint2USP(CurrentPointIndex);

        if (FoundTriangle2 == NONE)
          SplitOneTriangle(FoundTriangle1);
        else
          SplitTwoTriangles(FoundTriangle1, FoundTriangle2);
      }
      else
        TheClosestIndex = FindTheNextClosestPoint(TheClosestIndex);

      if (TheClosestIndex == NONE)
      { 
        Points[CurrentPointIndex].TriangulatedFlag = false;
        goto ven; 
      }
    }
  }
  else
  { int testtest = 0;
  }
ven: ;

}

void Triangulate()
{ InsertTheFirstPoint();

FILE *f;

  for (long i = 1; i < OriginalNoOfVertices; i++)
  {
 /*   f = fopen("napaka.err", "wt");
    fprintf(f, "%ld \n", i);
    fclose(f);
*/
   InsertPoint(i);
  }
}




void PlotTriangulatedPoints(HDC hdc, double BoxXMin, double BoxYMin, 
                            double BoxXMax, double BoxYMax, int r_right, int r_bottom,
                            int NoPts)
{ 
    int xi, yi;

    for (long i = 0; i < NoPts; i++)
    { if (Points[i].TriangulatedFlag)
      { xi = FromRealToPixel(Points[i].x, BoxXMin, BoxXMax - BoxXMin , r_right);
        yi = r_bottom - FromRealToPixel(Points[i].y, BoxYMin, BoxYMax - BoxYMin , r_bottom);

        MoveToEx(hdc, xi-3, yi, NULL);
        LineTo(hdc, xi+3, yi);

        MoveToEx(hdc, xi, yi-3, NULL);
        LineTo(hdc, xi, yi+3);
      }
    }
 }



void Triangulate(HDC hdc, double BoxXMin, double BoxYMin, 
                          double BoxXMax, double BoxYMax, 
                          int r_right, int r_bottom)
{
// FILE* ff; 

  InsertTheFirstPoint();
  for (long i = 1; i < OriginalNoOfVertices; i++)
  { 

// ff = fopen("1.err", "wt");
//fprintf(ff, "%ld \n", i);
//fclose(ff);

   PlotPoints(hdc, BoxXMin, BoxYMin, BoxXMax, BoxYMax, r_right, r_bottom, i+1);


if (NextTrianglePosition > 79)
{int oooo = 1;}

    if (i == 25) // 90
    {  SosediPravi();
       int ooooo = 1;
    }
    InsertPoint(i);

    if (SosediPravi() != 2)
    { int evoproblema = 1; }


    PlotTriangles(hdc, BoxXMin, BoxYMin, BoxXMax, BoxYMax, r_right, r_bottom);
    PlotPoints(hdc, BoxXMin, BoxYMin, BoxXMax, BoxYMax, r_right, r_bottom, i+1);
  }

}


//// auxilliar

void PrintTrinagles(char* fname)
{ long i;

  FILE* f = fopen(fname, "wt");

  for (i = 0; i < NextTrianglePosition; i++)
    fprintf(f, "No: %ld,  i=%ld j=%ld k=%ld,   ij=%ld jk=%ld ki=%ld \n",
                 i, TriangleArray[i].i, TriangleArray[i].j, TriangleArray[i].k,
                 TriangleArray[i].ij, TriangleArray[i].jk, TriangleArray[i].ki);
  

  fclose(f);
}


void PlotTriangles(HDC hdc, double BoxXMin, double BoxYMin, double BoxXMax, double BoxYMax, 
                            int right, int bottom)
{ long ii;
  double x1, x2, x3, y1, y2, y3;
  int xi1, yi1, xi2, yi2, xi3, yi3;



  POINT plxy[5];

  HPEN hpen, hpenOld;
  HBRUSH HBrush = CreateSolidBrush(RGB(0,255,0));
  (HBRUSH)SelectObject(hdc, HBrush);

  hpen = CreatePen(PS_SOLID, 1, RGB(255,0,0));
  hpenOld = (HPEN)SelectObject(hdc, hpen);


  for (ii = 0; ii < NextTrianglePosition; ii++)
  {
    if ( (TriangleArray[ii].i < OriginalNoOfVertices) &&
         (TriangleArray[ii].j < OriginalNoOfVertices) &&
         (TriangleArray[ii].k < OriginalNoOfVertices) )
    { x1 = Points[TriangleArray[ii].i].x;
      y1 = Points[TriangleArray[ii].i].y;

      x2 = Points[TriangleArray[ii].j].x;
      y2 = Points[TriangleArray[ii].j].y;

      x3 = Points[TriangleArray[ii].k].x;
      y3 = Points[TriangleArray[ii].k].y;

      xi1 = FromRealToPixelX(x1, BoxXMin, BoxXMax - BoxXMin, right);
      yi1 = bottom - FromRealToPixelY(y1, BoxYMin, BoxYMax - BoxYMin, bottom);
      xi2 = FromRealToPixelX(x2, BoxXMin, BoxXMax - BoxXMin, right);
      yi2 = bottom - FromRealToPixelY(y2, BoxYMin, BoxYMax - BoxYMin, bottom);
      xi3 = FromRealToPixelX(x3, BoxXMin, BoxXMax - BoxXMin, right);
      yi3 = bottom - FromRealToPixelY(y3, BoxYMin, BoxYMax - BoxYMin, bottom);

      plxy[0].x = xi1;   plxy[0].y = yi1;
      plxy[1].x = xi2;   plxy[1].y = yi2;
      plxy[2].x = xi3;   plxy[2].y = yi3;

      Polygon(hdc, plxy, 3);
 
      MoveToEx(hdc, xi1, yi1, NULL);
      LineTo(hdc, xi2, yi2);
      LineTo(hdc, xi3, yi3);
      LineTo(hdc, xi1, yi1);
    }
  }


  SelectObject(hdc, hpenOld);
  DeleteObject(hpen);
}




void PlotPoints(HDC hdc, double BoxXMin, double BoxYMin, double BoxXMax, double BoxYMax, 
                         int r_right, int r_bottom, long ii)
{ 
    int xi, yi;

    for (long i = 0; i < ii; i++)
    { 
      xi = FromRealToPixel(Points[i].x, BoxXMin, BoxXMax - BoxXMin , r_right);
      yi = r_bottom - FromRealToPixel(Points[i].y, BoxYMin, BoxYMax - BoxYMin , r_bottom);

      MoveToEx(hdc, xi-3, yi, NULL);
      LineTo(hdc, xi+3, yi);

      MoveToEx(hdc, xi, yi-3, NULL);
      LineTo(hdc, xi, yi+3);

    }
 }




void  TestClosestPoint()
{ long K;
  long L;

  InsertPoint2USP(0);
  for (long i = 1; i < OriginalNoOfVertices; i++)
  { 
 
 
    K = DetermineTheClosestPoint(i);
    L = TheClosestPointBruteForce(i);
    
    if (K != L)
    {  double a1 = Points[L].x - Points[i].x;
       double b1 = Points[L].y - Points[i].y;
       double d1 = a1 * a1 + b1 * b1;

       double a2 = Points[K].x - Points[i].x;
       double b2 = Points[K].y - Points[i].y;
       double d2 = a2 * a2 + b2 * b2;

       if (Equal(d1, d2, MyEpsilon) == 0)
         int qqqq = 0;
    }

    InsertPoint2USP(i);
  }

}


long TheClosestPointBruteForce(long k)
{ int i;
  double d = MAX_DOUBLE; 
  double d1;
  long iii;
  double a, b;

  for (i = 0; i < k; i++)
  { a = Points[i].x - Points[k].x;
    b = Points[i].y - Points[k].y;
    d1 = a * a + b * b;
    if (d1 < d)
    { d = d1;
      iii = i;
    }
  }
  return iii;
}

long* TheNextClosestPointBruteForce(long k)
{

  double* DistArr = new double[k + 10];
  long* IndArr = new long[k + 10];
  long i;

  for (i = 0; i < k; i++)
  { DistArr[i] = (Points[i].x - Points[k].x) * (Points[i].x - Points[k].x) + 
                 (Points[i].y - Points[k].y) * (Points[i].y - Points[k].y);
    IndArr[i] = i;
  }

  long j;
  double t;
  long ti;

  for (i = 0; i < k-1; i++)
    for (j = i+1; j < k; j++)
    { if (DistArr[i] > DistArr[j])
      { t = DistArr[i];
        ti = IndArr[i];
        DistArr[i] = DistArr[j];
        IndArr[i] = IndArr[j];
        DistArr[j] = t;
        IndArr[j] = ti;
      }
    }

  delete[] DistArr;

  return IndArr;
}


long ReturnNoOfLegalisations() {return NoOfLegalisations;}


void  TestTheNextClosestPoint()
{ long K, K1;
  long L;

//  FILE* f1;

  InsertPoint2USP(0);
  for (long i = 1; i < OriginalNoOfVertices; i++)
  { 
  
//    f1 = fopen("prox.txt", "wt");
//    fprintf(f1, "%ld \n", i);
//    fclose(f1);

    if (i == 5)
    {
int ooo = 0;
    }

 
 
    K = DetermineTheClosestPoint(i);
    


    long* LArr = TheNextClosestPointBruteForce(i);
    
    for (long u= 1; u < i; u++)
    { K1 = ReturnTheNextClosestPoint();
      L = LArr[u];

    if (K1 != L)
    {  double a1 = Points[L].x - Points[i].x;
       double b1 = Points[L].y - Points[i].y;
       double d1 = a1 * a1 + b1 * b1;

       double a2 = Points[K1].x - Points[i].x;
       double b2 = Points[K1].y - Points[i].y;
       double d2 = a2 * a2 + b2 * b2;

       if (Equal(d1, d2, MyEpsilon) == 0)
         int qqqq = 0;
    }

    }

    delete[] LArr;

    InsertPoint2USP(i);
  }

}
