// USPTriangF.h: interface for the USPTriangF class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_USPTRIANGF_H__B95CE2C2_5C62_11D5_9B3F_004F4902DA76__INCLUDED_)
#define AFX_USPTRIANGF_H__B95CE2C2_5C62_11D5_9B3F_004F4902DA76__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Proximity2L.h"
#include "utility.h"

#define NONE -1
#define NO_OF_SURROUNDING_TRIANGLES 30
 #define START_2USP 500  // optimal
//#define START_2USP 0
#define THE_SAME_POINT -100
#define NO_MORE_TRIANGLES -1


typedef struct ListOfTrianglesType1     //dinamical list of integers
{ long TriangleIndex;                   // for triangle indices
  struct ListOfTrianglesType1 *next;
} ListOfTrianglesType;


// statical & dinamical list of integers of triangles surrounding 
// a vertex. The NO_OF_SURROUNDING_TRIANGLES positions is reserved
// for statical table. If there is more surrounding triangles,
// these are stored in a dinamical list.
typedef struct SurroundingTriangleInfo   
{ bool FlagsArray[NO_OF_SURROUNDING_TRIANGLES];      // if false, position is not used
  long SurroundingArray[NO_OF_SURROUNDING_TRIANGLES];// list of triangles
  int NoT;                                           // No of triangles
  ListOfTrianglesType* ListofTriangles;
} SurroundingTriangleInfo;


typedef struct MyPointT1// point to be triangulated
{ long Index;
  double x, y;
  int CloseCheckFlag;
  bool TriangulatedFlag;
  SurroundingTriangleInfo ListOfTriangles; 
} MyPointT;


typedef struct bt1
{

  long i,j,k;                    // triangle coordinate
  long ij,     // pointers to triangle neighbours
       ki,
       jk;
  bool Eij, Eki, Ejk;             // true if has been legalised already;
  bool BorderTriangleFlag;         // for sweep-line algorithm
} BaseTriangle;


// internal functions
void FinalUSPTriang();
void AddTriangleToPoint(long index, long t1);
void DeleteTriangleFromPoint(long index, long t1);
int PointInTriangle(long TheClosestIndex, long& t1, long& t2);
long ReturnTriangle(long Position);
int ExamineTriangle(double tpx, double tpy, long index, long &t1, long &t2);
bool PointOnEdge(double x1, double y1, double x2, double y2, 
                 double tpx, double tpy, double& a1);
void SplitOneTriangle(long OldT);
void SplitTwoTriangles(long OldT1, long OldT2);
void ReplaceNeighbour(long TIndex, long OldT, long NewNeighbour);
void ReplaceNeighbours(long TIndex, long OldT1, long NewT1, long OldT2, long NewT2);

int AdjustVertices(long t1, long t2);
void ShiftVerticesRight(BaseTriangle* t);
void Legalise(long t1, long t2);
long FindTheNextClosestPoint(long Previous);

/////  outher functions
void InitUSPTriang(long NoOfPoints, double* X,  double* Y,
                   double BXMin, double BYMin, 
                   double BXMax, double BYMax);

void Triangulate();


void InsertTheFirstPoint();
void InsertPoint(long i);
void PrintTrinagles(char*);

void PlotTriangles(HDC hdc, double BoxXMin, double BoxYMin, double BoxXMax, double BoxYMax, 
                            int right, int bottom);

void PlotTriangulatedPoints(HDC hdc, double BoxXMin, double BoxYMin, 
                            double BoxXMax, double BoxYMax, int r_right, int r_bottom,
                            int NoPts);

void  TestClosestPoint();
void  TestTheNextClosestPoint();
long TheClosestPointBruteForce(long k);
long* TheNextClosestPointBruteForce(long k);

void Triangulate(HDC hdc, double BoxXMin, double BoxYMin, 
                          double BoxXMax, double BoxYMax, 
                          int r_right, int r_bottom);


void PlotPoints(HDC hdc, double BoxXMin, double BoxYMin, double BoxXMax, double BoxYMax, 
                         int r_right, int r_bottom, long ii);


long ReturnPointInTriangleCounter(); 
long ReturnTheNextClosestPointCounter();
long ReturnNoOfLegalisations();

#endif // !defined(AFX_USPTRIANGF_H__B95CE2C2_5C62_11D5_9B3F_004F4902DA76__INCLUDED_)
