// ********** HEADER FILE ***********************

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <values.h>

class CPoint
{
  public:
    double x;
    double y;
};

class CValue
{
  public:
    CPoint p;
};

// DETERMINISTIC SKIP LIST
#define maxkey MAXDOUBLE
#define newnode (nodePtr)malloc(sizeof(struct node))

typedef struct node *nodePtr;
struct node
{
  double m_key;
  CValue m_value;
  nodePtr r;
  nodePtr d;
};

nodePtr               m_head;
nodePtr               m_bottom;
nodePtr               m_tail;



// ********** SOURCE FILE ***********************

// INICIALNIZACIJA
void InitDeterministicSkipList()
{
  m_head        = newnode;
  m_bottom      = newnode;
  m_tail        = newnode;
  m_head->m_key = maxkey;
  m_head->d     = m_bottom;
  m_head->r     = m_tail;
  m_bottom->r   = m_bottom;
  m_bottom->d   = m_bottom;
  m_tail->m_key = maxkey;
  m_tail->r     = m_tail;
}

// VSTAVLJANJE VSEH ELEMENTOV (TOCK V RAVNINI) V SKIP LIST
void MakeDeterministicSkipList()
{
  InitDeterministicSkipList();
  for (int i=0; i<m_inputSize; i++)
  {
    m_value.p = m_element[i];
    AddDeterministicSkipListElement(m_value.p.x);
  }
}

// ISKANJE VSEH ELEMENTOV SKIP LIST-A
void SearchDeterministicSkipList()
{
  for (int i=0; i<m_inputSize; i++)
  {
    m_value.p = m_element[i];
    SearchDeterministicSkipListElement();
  }
}

// BRISANJE VSEH ELEMENTOV SKIP LIST-A (LAHKO SE UPORABI KOT DESTRUKTOR)
void DeleteDeterministicSkipList()
{
  for (int i=0; i<m_inputSize; i++)
  {
    m_value.p = m_element[i];
    DeleteDeterministicSkipListElement();
  }
}

// VSTAVLJANJE NOVEGA ELEMENTA V SKIP LIST
void AddDeterministicSkipListElement(double key)
{
  nodePtr t,x;
  x = m_head;
  m_bottom->m_key   = key;
  while (x != m_bottom)
  {
    while (key > x->m_key)
      x = x->r;
    if (x->m_key > x->d->r->r->m_key)
    {
      t = newnode;
      t->r = x->r;
      t->d = x->d->r->r;
      x->r = t;
      t->m_key = x->m_key;
      t->m_value = x->m_value;
      x->m_key = x->d->r->m_key;
      x->m_value = x->d->r->m_value;
    }
    x = x->d;
  }
  if (m_head->r != m_tail)
  {
    t = newnode;
    t->d = m_head;
    t->r = m_tail;
    t->m_key = maxkey;
    m_head = t;
  }
}

// ISKANJE ELEMENTA SKIP LIST-A - V TEM PRIMERU JE TO 2D TOCKA
bool SearchDeterministicSkipListElement()
{
  double key = m_value.p.x;

  nodePtr x;
  x = m_head;
  m_bottom->m_key = key;
  while (key != x->m_key)
    x = (key < x->m_key) ? x->d : x->r;
  if (x != m_bottom)
    return true;
  else
    return false;
}

// BRISANJE ELEMENTA SKIP LIST-A - V TEM PRIMERU JE TO 2D TOCKA
void DeleteDeterministicSkipListElement()
{
  double key = m_value.p.x;

  nodePtr x,px,nx,t;
  double pred, lastAbove;

  x = m_head->d;
  m_bottom->m_key = key;
  lastAbove = m_head->m_key;
  while (x != m_bottom)
  {
    while (key > x->m_key)
    {
      px = x;
      x = x->r;
    }
    nx = x->d;
    if (x->m_key == nx->r->m_key)
    {
      if (x->m_key != lastAbove)
      {
        t = x->r;
        if ((t->m_key == t->d->r->m_key) || (nx == m_bottom))
        {
          x->r = t->r;
          x->m_key = t->m_key;
          x->m_value = t->m_value;
          free(t);
        }
        else
        {
          x->m_key = t->d->m_key;
          x->m_value = t->d->m_value;
          t->d = t->d->r;
        }
      }
      else
      if (px->m_key <= px->d->r->m_key)
      {
        if (nx == m_bottom)
          pred = px->m_key;
        px->r = x->r;
        px->m_key = x->m_key;
        px->m_value = x->m_value;
        free(x);
        x = px;
      }
      else
      {
        t = (px->m_key == px->d->r->r->m_key ? px->d->r : px->d->r->r);
        px->m_key = t->m_key;
        px->m_value = t->m_value;
        x->d = t->r;
      }
    }
    lastAbove = x->m_key;
    x = nx;
  }
  x = m_head->d;
  while (x != m_bottom)
  {
    while (key == x->m_key)
      x = x->r;
    if (key == x->m_key)
      x->m_key = pred;
    x = x->d;
  }
  if (m_head->d->r == m_tail)
  {
    x = m_head;
    m_head = x->d;
    free(x);
  }
}



