// Implementation of some small utilities.
//
// Written by B. ZALIK 
// Updated for polygon merging in April and May 19999
// e-mail: zalik@uni-mb.si
//////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "utility.h"
#include "globals.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#include <math.h>

int Equal(double a, double b)
{ if (ABS(a-b) <= EPSILON) return 1; else return 0; }

int Equal(double a, double b, double delta)
{ if (ABS(a-b) <= delta) return 1; else return 0; }

int FromRealToPixelX(double x, double xwmin, double dx, int pixelsize)
// converts real coordinate to pixel coordinate and returns it
// - x: coordinate which will be converted
// - xwmin: minimal value of the real window
// - dx: size of a real window
// - pixelsize: size of the window in pixels
{
  return ( (int) ( (double)((x - xwmin) * pixelsize) / dx) );
}

int FromRealToPixelY(double y, double ywmin, double dy, int pixelsize)
// converts real coordinate to pixel coordinate and returns it
// - y: coordinate which will be converted
// - ywmin: minimal value of the real window
// - dy: size of a real window
// - pixelsize: size of the window in pixels
{
  return (FromRealToPixelX(y, ywmin, dy, pixelsize));

//  return ( pixelsize - ((int) ( (double)((y - ywmin) * pixelsize) / dy)) );
}


int FromRealToPixel(double x, double xwmin, double dx, int pixelsize)
// converts real coordinate to pixel coordinate and returns it
// - x: coordinate which will be converted
// - xwmin: minimal value of the real window
// - dx: size of a real window
// - pixelsize: size of the window in pixels
// The function can be aplied for X and Y direction
{
  return ( (int) ( (float)((x - xwmin) * pixelsize) / dx) );
}

int Signum(double a)
{ if (Equal(a, 0) ) return 0;
  if (a > 0) return 1;
  return -1;
}


int Sign(double a)
{ if (a >=0) return 1;
  return -1;
}

int Sign(int a)
{ if (a >=0) return 1;
    return -1;
}

int Potenca(int osnova, int a)
{ if (a == 0) return 1;

  int i;
  int result = 1;
  for (i = 0; i < a; i++)
	 result *= osnova;

  return result;
}


int Odd(int k)
{ return (k % 2);
}

int Normalize(Point& v)
{ double norma = sqrt(v.x * v.x + v.y * v.y);
 
  if (norma == 0) 
    return 0;
  v.x = v.x / norma;
  v.y = v.y / norma;
  return 1;
}


double Sqr(double a)
{  return (a*a);
}

void GetVector(Point3 &a, Point3 &b, Point3& c)
{ c.x = b.x - a.x;
  c.y = b.y - a.y;
  c.z = b.z - a.z;
}

void VectorProduct(Point3& a, Point3& b, Point3& out)
{ out.x = a.y * b.z - a.z * b.y;
  out.y = a.z * b.x - a.x * b.z;
  out.z = a.x * b.y - a.y * b.x;
}

double VectorProduct(Point& a, Point& b)
{ return (a.x * b.y - a.y * b.x);
}


int TheSameSign(Point3& p1, Point3& p2, Point3& q1, Point3& q2)
// fucntion determines if the line q1q2 lies in the same half plane determined
// by line p1p2. If it does, the the function returns 1
{ Point3 a, b, c;
 
  GetVector(p2, p1, a);
  GetVector(q1, p1, b);
  GetVector(q2, p1, c);

  Point3 r1, r2;
  VectorProduct(a, b, r1);
  VectorProduct(a, c, r2);

  if ((Equal(r1.z, 0) == 1) || (Equal(r2.z, 0) == 1)) // lines touches
    return 0;

  if ((r1.z <= 0) && (r2.z <= 0)) return 1;
  if ((r1.z >= 0) && (r2.z >= 0)) return 1;

  return 0;
}

double ScalarProduct(Point3 &a, Point3 &b)
{ return(a.x * b.x + a.y * b.y + a.z * b.z); }


double ScalarProduct(double x1, double y1, double x2, double y2,
                     double x3, double y3)
{ double ax = x1 - x2;
  double bx = x3 - x2;
  double ay = y1 - y2;
  double by = y3 - y2;

  return(ax * bx + ay * by); 
}

double LenghtOfVector(Point3 &a)
{ return (sqrt(SQR(a.x) + SQR(a.y) + SQR(a.z))); }


int Collinear2D(double x1, double y1, double x2, double y2, double x3, double y3)
{ Point3 p1 = {x1, y1, 0};
  Point3 p2 = {x2, y2, 0};
  Point3 p3 = {x3, y3, 0};

  Point3 v1, v2, r;

  GetVector(p1, p2, v1);
  GetVector(p1, p3, v2);
  VectorProduct(v1, v2, r);
  if (ABS(r.z) <= 0.01 ) return 1;
  return 0;
}

double ReturnTangAgainstX(double x, double y, int &quadrant)
// function calculates and tng alpha of a vector against X axis. In quadrant
// int returns a code of quadrant in which the vector lies
// x, y are the end points of the vector (it must be translated into origin
{ if (Equal(x, 0) == 1)
  { if (y > 0)
	 { quadrant = 1;
		return PI_HALF;
	 }
	 else
	 { quadrant = 3;
		return (PI + PI_HALF);
	 }
  }
  else
  { if (x > 0)
    { if (Equal(y, 0) == 1)
      { quadrant = 1;
        return 0;
      } 
      else 
		if (y > 0) quadrant = 1;
		else quadrant = 4;
    }
	else
    { if (Equal(y, 0) == 1)
      { quadrant = 3;
        return PI;
      }
      else
		if (y > 0) quadrant = 2;
		else quadrant = 3;
     }
	 return ( ABS(atan( (double)(y / x) ) ) );
  }
}

double AngleAgainstX(double x, double y)
// Function calculates an angle between a local vector (x,y) and X coordinate
// returns angle in radians
{ int q;
  double tan = ReturnTangAgainstX(x, y, q);

  if ((Equal(x, 0) == 1) || (Equal(y, 0) == 1))
    return tan;

  switch (q)
  { case 1: return tan;
	 case 2: return (PI - tan);
	 case 3: return (PI + tan);
	 case 4: return (2 * PI - tan);
  }
  return 0.0;
}



double DistanceBetween2DPoints(double x1, double y1, double x2, double y2)
{ double a = x2 - x1;
  double b = y2 - y1;

  return (sqrt (a * a + b * b) );
}



int Between0_1(double a)
{ if (( a > 0 ) && ( a < 1.0)) return 1;
  return 0;
}



double Maximum(double a, double b)
{ if (a >= b) return a;
  return b;
}

double Minimum(double a, double b)
{ if (a <= b) return a;
  return b;
}



int TheSamePoints2D(double x1, double y1, double x2, double y2, double delta)
{ if ((Equal(x1, x2, delta) == 1) && (Equal(y1, y2, delta) == 1)) return 1;
  return 0;
}


int TheSamePoints2D(double x1, double y1, double x2, double y2)
{
  return TheSamePoints2D(x1, y1, x2, y2, EPSILON);
}


int TheSamePoints2D(Point* p1, Point* p2)
{ return TheSamePoints2D(p1->x, p1->y, p2->x, p2->y, EPSILON); }

int TheSamePoints2D(Point* p1, Point* p2, double delta)
{ return TheSamePoints2D(p1->x, p1->y, p2->x, p2->y, delta); }

void Swap(int& a, int& b)
{ int c = a;
  a = b;
  b = c;
}

void Swap(double& a, double& b)
{ double c = a;
  a = b;
  b = c;
}

void Swap(Point& a, Point& b)
{ Point c = a;
  a = b;
  b = c;
}


int IsDecimalNumber(double u)
// returns 1 if u is not a number close enough to and integer
// if it is integer then it returns 0
{ if (Equal(((int)u), u) == 1) return 0;
  return 1;
}


int BigEndianLongInt(int a)
// function converts the Big-endian byte alingment (useon on Unix) to
// the Little-endian byte alignment and returns it. It works for 
// 32-bits long integers !!!!!
// for example: if a = 0x12345678
// it returns 0x78563412
{ int byte1 = a & 0xFF000000;
  int byte2 = a & 0x00FF0000;
  int byte3 = a & 0x0000FF00;
  int byte4 = a & 0x000000FF;
  byte1 = byte1 >> 24;
  byte1 = byte1 & 0x000000FF;
  byte2 = byte2 >> 8;
  byte2 = byte2 & 0x0000FF00;
  byte3 = byte3 << 8;
  byte3 = byte3 & 0x00FF0000;
  byte4 = byte4 << 24;
  byte4 = byte4 & 0xFF000000;

  int bei = byte1;
  bei = bei ^ byte2;
  bei = bei ^ byte3;
  bei = bei ^ byte4;

  return bei;
}


int SquareEquation(double a, double b, double c, double& xout1, double& xout2)
{ double det = b * b - 4 * a * c;

  if (Equal(det, 0) == 1) // one solution
  { xout1 = xout2 = -b / (2 * a);
    return 1;
  }

  if (det < 0) // no real solutions
    return 0;

  // two real solutions
  double a2 = 2* a;
  double sq = sqrt(det);
  xout1 = (-b + sq)/ a2;
  xout2 = (-b - sq)/ a2;
  return 2;
}

/*
double SignedArea(double x1, double y1, double x2, double y2, double x3, double y3)
{ return (x1 * (y2 - y3) - y1 * (x2 - x3) + (x2 * y3 - y2 * x3)); }
*/

double ABSOL(double a)
{ if (a >= 0) return a;
  return -a;
}

double SignedArea(double x1, double y1, double x2, double y2, double x3, double y3)
{ double a;


  if (Equal(y1, y2) && Equal(y2, y3))
    return 0;

  a = x1 * (y2 - y3) - y1 * (x2 - x3) + (x2 * y3 - y2 * x3); 
  return a;
}



int TriangleOutscribedCircle(double x1, double y1, double x2, double y2, 
                             double x3, double y3, double& x, double& y, double& rr)
// method calculates the centre and the polmer of the cricle determined by three points
// if possible. IMPORTANT; it returns the sqared value of the polmer.
// if neccessary, this value has to be rooted outside this routine
{ double d = (y1 - y3) * (x1 - x2) - (y1 - y2) * (x1 - x3);

  if (Equal(d ,0) == 1) return 0;  // points are colinear

  double xx1 = x1 * x1;
  double yy1 = y1 * y1;
  double xx2 = x2 * x2;
  double yy2 = y2 * y2;
  double xx3 = x3 * x3;
  double yy3 = y3 * y3;

  double a = xx1 + yy1 - xx2 - yy2;
  double b = xx1 + yy1 - xx3 - yy3;

  x = 0.5 * (a * (y1 - y3) - b * (y1 - y2)) / d;
  y = 0.5 * (b * (x1 - x2) - a * (x1 - x3)) / d;

  double f = (x - x1);
  f = f * f;
  double g = (y - y1);
  g = g * g;

  rr = f + g;

  return 1;
}

